import os, sys 
current_directory = os.path.dirname(os.path.realpath(__file__))
sys.path.append(current_directory)

import re
from utils import OPERATOR, FUNCTION, find_functions, find_calculates, find_tag_attrs, find_orderby


def find_temps(text):
    if not re.findall(r'[\w`]+\.[\w`]+', text):
        return ['temp']
    else:
        return []



def find_limit(text):
    match = re.findall(r'\blimit\s+(\d+)', text)
    if match:
        limit_value = [int(i) for i in match]
        # print(limit_value)
        return limit_value
    else:
        return []


# def find_orderby(text):
#     # print(text)
#     word_list = text.split(' ')
#     order_index = word_list.index('order')
#     try:
#         order_attr = '.'.join(word_list[order_index - 1].split('.')[1:])
#         # print('order_attr', order_attr)
#         # order_attr_var1 = word_list[order_index - 1]
#         # order_attr_var2 = word_list[order_index + 2]
#         try:
#             order_sort = word_list[order_index + 3]
#         except:
#             order_sort = 'asc'
#         orderby = (order_attr, order_sort)
#         return orderby
#     except:
#         try:
#             order_sort = word_list[order_index + 2]
#         except:
#             order_sort = 'asc'
#         orderby = (word_list[order_index + 2], order_sort)
#         return orderby


def has_english_characters(s):
    return bool(re.search(r'[a-zA-Z]', s))


def is_numeric_using_regex(s):
    return bool(re.match(r'^-?\d+(\.\d+)?$', s))




def find_returns(text):
    text = re.sub(r'\s+', ' ', text)  # 去除多个连续空格
    text = text.lower().replace('return ', '')  # 去除return字符
    text = re.sub(r'\s+as\s+\w+\s*', ' ', text)  # 去除自命名
    rag_attrs = []
    calculates = []
    functions = []
    orderbys = []
    temps = []
    limits = []

    # 首先找到limit
    if ' limit ' in text:
        match = re.search(r'\blimit\s+(\d+)\s*(.*)$', text)
        if match:
            limit = match.group(1).strip()
            limits.append(limit)
            # print(limit)
            text = re.sub(r'\s*limit\s+\d+\s*.*$', '', text)
        else:
            pass

    # print(text)
    sents = text.split(',')
    sents = [i.strip() for i in sents]
    # print(sents)

    for sent in sents:
        # print(sent)
        # 涉及orderby
        if 'order by' in sent:
            orderby = find_orderby(sent)
            orderbys.append(orderby)
            # print(orderby)
            continue
        # 涉及计算
        if any(substring in sent for substring in [' ' + i + '' for i in OPERATOR]):
            # print(0)
            calculate = find_calculates(sent)
            calculates.append(calculate)
            continue
        if any(substring in sent for substring in [i + '(' for i in FUNCTION]):
            function = find_functions(sent)
            functions.append(function)
            continue
        # 定义temp
        if not re.findall(r'[\w`]+\.[\w`]+', text):
            temps.append('temp')
            continue
        rag_attr = find_tag_attrs(sent)
        rag_attrs.append(rag_attr)
        # print(rag_attr)
    splits = {
        'return_rag_attrs': sorted(rag_attrs),
        'return_calculates': sorted(calculates),
        'return_functions': sorted(functions),
        'return_orderby': sorted(orderbys),
        'return_temps': sorted(temps),
        'return_limit': sorted(limits),
    }
    return splits


def return_evaluate(gold, pre):
    # math.isclose(10.00, 10)
    splits_gold = find_returns(gold)
    splits_pre = find_returns(pre)
    for key in splits_gold.keys():
        if splits_gold[key] != splits_pre[key]:
            return False
    return True


if __name__ == '__main__':
    gold = 'RETURN sd.stock_data.`date`, sd.stock_data.closing_price, h.position_ratio'
    gold = 'RETURN pof.public_offering_fund.name , pof.public_offering_fund.net_worth, pof.public_offering_fund.management_fee LIMIT 50'
    # gold = 'RETURN sd.stock_data.closing_price , sd.stock_data.opening_price , avg(td.trade_data.closing_price) AS avg_closing_price'
    gold = 'RETURN distinct(sd.stock_data.closing_price), sd.stock_data.maximum_price as SSDCP ORDER BY SSDCP ASC LIMIT 1'
    # gold = 'RETURN (sd.stock_data.closing_price - p1) / p1 as sum_range'
    # gold = 'RETURN pof.public_offering_fund.management_fee > 10.00'
    # gold = 'RETURN temp'

    pre = 'RETURN sd.stock_data.`date`, h.position_ratio, sd.stock_data.closing_price'
    pre = 'RETURN pof.public_offering_fund.name , pof.public_offering_fund.net_worth, pof.public_offering_fund.management_fee LIMIT 50'
    # pre = 'RETURN sd.stock_data.closing_price , avg(dgsdf.trade_data.closing_price) , sd.stock_data.opening_price'
    pre = 'RETURN sd.stock_data.closing_price, ORDER BY sd.stock_data.maximum_price   LIMIT 1'
    # pre = 'RETURN (sd.stock_data.closing_price - ddd) / ddd as s'
    # pre = 'RETURN avg(pof.public_offering_fund.management_fee) > 10'
    # pre = 'RETURN temp1'

    splits = find_returns(gold)
    print(splits)
    splits = find_returns(pre)
    print(splits)

    result = return_evaluate(gold, pre)
    print(result)




from pygments.token import *
from antlr4.tree.Tree import TerminalNodeImpl
from antlr4 import InputStream, CommonTokenStream

from ..grammar.SparqlParser import SparqlParser
from ..grammar.SparqlLexer import SparqlLexer

from ...utils import parse_layer

def parse_sparql(sparql, tree_only=False):
    """ 返回输入sparql语句的AST树 
        AST 树、序列化的树
        便利后得到的 每个token 及其对应的父节点
    """
    input_stream = InputStream(sparql)
    # print(input_stream)
    lexer = SparqlLexer(input_stream)
    stream = CommonTokenStream(lexer)
    parser = SparqlParser(stream)
    tree = parser.query()
    tree.toStringTree()
    if tree_only:
        return tree, parser  
        
    parts, parents = [], []
    # print(tree)
    # traverse(tree, parser.ruleNames)
    parse_layer(tree, parser.ruleNames, parts, parents)
    print("after parsing", tree)
    for token, par in zip(parts, parents):
        print(token, "\t", par)
    s = tree.toStringTree(recog=parser)
    return tree, s, parts, parents 

def get_sparql_nodes(tree, parser):
    """遍历语法树 返回树字符串、节点列表、  

    Args:
        tree: 
        parser: 
    """
    parts, parents = [], []
    parse_layer(tree, parser.ruleNames, parts, parents)
    s = tree.toStringTree(recog=parser)
    return s, parts, parents 

def split_sparql(query, lexer=None):
    """使用lexer将查询语句分词（语义单元级别

    Args:
        query: 查询字符串，可以包含语法错误
        lexer: pygments的lexer. Defaults to cy_lexer.

    Returns:
        a list of strings 
    """
    if lexer is None:
        try:
            from pygments.lexers.rdf import SparqlLexer as SPLexer
            lexer = SPLexer()
        except:
            raise Exception(f"lexer is not set!")

    pred_units = []
    for tag, substr in lexer.get_tokens(query):
        if tag is Token.Error:
            # return -1
            pred_units.append("错")
        # print(tag, substr)
        if substr.strip() != "":
            pred_units.append(substr)

    return pred_units




import os, sys 
current_directory = os.path.dirname(os.path.realpath(__file__))
sys.path.append(current_directory)
import re
from utils import OPERATOR, UNDERAND, find_calculates, remove_quotes, add_spaces_around_operators


def find_underand(text):
    text = remove_quotes(text)
    # print('text', text)
    for i in UNDERAND:
        if ' ' + i + '' in text:
            contents = []
            sents = text.split(i)
            sents = [i.strip() for i in sents]
            # print('sents', sents)
            for sent in sents:
                content = find_calculates(sent)
                contents.append(content)
            return (i, sorted(contents))
    return ''


def find_wheres(text):
    calculates = []
    temps = []
    # text = add_spaces_around_operators(text)
    text = re.sub(r'\s+', ' ', text)  # 去除多个连续空格
    text = text.lower().replace('where ', '')  # 去除where字符
    # print(text)
    sents = text.split('and')
    sents = [i.strip() for i in sents]
    # print(sents)
    for sent in sents:
        # print('sent', sent)
        if any(substring in sent for substring in [' ' + i + '' for i in UNDERAND]):
            calculate = find_underand(sent)
            # print(calculate)
            calculates.append(calculate)
            continue
        if any(substring in sent for substring in [' ' + i + '' for i in OPERATOR]):
            # print(sent)
            calculate = find_calculates(sent)
            # print(calculate)
            calculates.append(calculate)
            # print('calculate', calculate)
            continue
        if not re.findall(r'[\w`]+\.[\w`]+', text):
            temps.append('temp')
            continue

    splits = {
        'where_calculates': sorted(calculates, key=lambda x: x[0] if isinstance(x, tuple) else x),
        'where_temps': sorted(temps)
    }
    return splits


def where_evaluate(gold, pre):
    splits_gold = find_wheres(gold)
    splits_pre = find_wheres(pre)
    for key in splits_gold.keys():
        if splits_gold[key] != splits_pre[key]:
            return False
    return True


if __name__ == '__main__':
    gold = "WHERE s.stock.name == '[s]'"
    gold = "WHERE s.stock.name == '天力锂能' AND sd.stock_data.`date` == date('2023-8-31')"
    splits = find_wheres(gold)
    print(splits)



# -*- coding: utf-8 -*-
# @File    :   cli.py
# @Time    :   2023/09/23 20:22:19
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
用于实现qinit命令
'''

import os
import sys 
import shutil
from argparse import ArgumentParser


def copy_to_cwd(source):
    """ 
        将source目录下的所有内容复制到cwd目录中
        等价于 cp -r source/* ./

        TODO: 实际build python包的时候，空目录不被包含在内
    """
    for file in os.listdir(source):
        
        abs_path = os.path.join(source, file)
        if os.path.isfile(abs_path):
            dst_file = os.path.join(os.getcwd(), file)
            shutil.copyfile(abs_path, dst_file)
        if os.path.isdir(abs_path):
            dst_dir = os.path.join(os.getcwd(), file)
            shutil.copytree(abs_path, dst_dir)
 

def main():
    """
        在当前目录init
        qinit 
        qinit .
        -----------
        在指定目录init
        qinit new_project
        qinit --dirname new_project
    """
    parser = ArgumentParser()
    parser.add_argument('dirname', nargs="?", type=str, default=None, help='positional argument')
    parser.add_argument('--dirname', dest="opt_dirname", type=str, default=None, help='')
    args = parser.parse_args()
    # 获取模板目录的路径
    template_dir = os.path.join(os.path.dirname(__file__), 'code_template')
    # print(args.dirname, args.opt_dirname)
    # exit()
    if args.opt_dirname is None :
        if args.dirname is None or args.dirname == '.' :
            copy_to_cwd(template_dir)
        else:
            shutil.copytree(template_dir, args.dirname)
    else:
        shutil.copytree(template_dir, args.opt_dirname)

if __name__ == '__main__':
    sys.exit(main())   



from typing import * 
import torch 
from qdls.data import sequence_padding

from ..register import registers

@registers.collator.register("causal_collator")
class CausalCollator:
    """ 
        训练时right padding，测试时left padding
        fit for all causal language model
    """
    def __init__(self, tokenizer, mode='train') -> None:
        self.tokenizer = tokenizer
        self.mode = mode
        assert self.tokenizer.pad_token_id is not None, "tokenizer must have pad_token_id"

    def __call__(self, features) -> Any:
        if self.mode == 'test':
            input_ids= sequence_padding([x['input_ids'] for x in features ], padding=self.tokenizer.pad_token_id, padding_side='left')
            attention_mask= sequence_padding([ x['attention_mask'] for x in features ], padding=0, padding_side='left')
            labels = None 
        else:
            input_ids= sequence_padding([x['input_ids'] for x in features ], padding=self.tokenizer.pad_token_id)   #  
            attention_mask= sequence_padding([ x['attention_mask'] for x in features ])
            labels = sequence_padding([x['labels'] for x in features ], padding=self.tokenizer.pad_token_id)

        return {
            'input_ids' : torch.tensor(input_ids).long() ,                  # collate时的截断需要重新考虑 TODO
            'attention_mask' : torch.tensor(attention_mask).long(),
            'labels': torch.tensor(labels).long() if labels is not None else None ,
        }
        

@registers.collator.register("seq2seq_collator")
class Seq2seqCollator:
    """ 
        T5, Bart 等seq2seq模型的collator
    """
    def __init__(self, tokenizer, mode='train') -> None:
        self.tokenizer = tokenizer
        self.mode = mode

    def __call__(self, features) -> Any:
        if self.mode == 'test':
            input_ids= sequence_padding([x['input_ids'] for x in features ], padding=self.tokenizer.pad_token_id)
            attention_mask= sequence_padding([ x['attention_mask'] for x in features ], padding=0)
            labels = None 
        else:
            input_ids= sequence_padding([x['input_ids'] for x in features ], padding=self.tokenizer.pad_token_id)   #  
            attention_mask= sequence_padding([ x['attention_mask'] for x in features ], padding=0)
            labels = sequence_padding([x['labels'] for x in features ], padding=self.tokenizer.pad_token_id)

        return {
            'input_ids' : torch.tensor(input_ids).long(),
            'attention_mask' : torch.tensor(attention_mask).long(),
            'labels': torch.tensor(labels).long() if labels is not None else None ,
        }
    



import re
from datetime import datetime

OPERATOR = ['==', '!=', '<>', '>', '>=', '<', '<=', 'is null', 'is not null', 'is empty', 'is not empty',
            '+', '-', '*', '/', '%', '!',
            'and', 'or', 'not', 'xor',
            'union', 'union all', 'intersect', 'minus',
            'contains', 'in', 'starts with', 'ends with', '', '']
FUNCTION = ['abs', 'floor', 'ceil', 'round', 'sqrt', 'cbrt', 'hypot', 'pow', 'exp', 'exp2', 'log', 'log2', 'log10',
            'sin', 'asin', 'cos', 'acos', 'tan', 'atan', 'rand', 'rand32', 'rand64', 'size', 'range', 'sign',
            'avg', 'count', 'max', 'min', 'collect', 'std', 'sum',
            'date',
            'distinct']
UNDERAND = ['or', 'not', 'xor']
SPLIT = ['match', 'where', 'with', 'return']


def has_english_characters(s):
    return bool(re.search(r'[a-zA-Z]', s))


def is_numeric_using_regex(s):
    return bool(re.match(r'^-?\d+(\.\d+)?$', s))


# def is_valid_date_string(date_str):
#     print('date_str', date_str)
#     date_str = remove_quotes(date_str)
#     date_object = datetime.strptime(date_str, "%Y-%m-%d")
#     formatted_date_string = date_object.strftime("%Y-%m-%d")
#     # pattern = re.compile(r'\b\d{4}-\d{2}-\d{2}\b')
#     match = pattern.match(date_str)
#     return bool(match)


def add_spaces_around_operators(sentence):
    for op in ['<=', '>=', '==', '<', '>', '+']:
        pattern = re.compile(rf'(?<=[^\s]){re.escape(op)}(?=[^\s])')
        sentence = pattern.sub(f' {op} ', sentence)
    return sentence


def is_datetime_format(input_string, date_formats=["%Y-%m-%d", "%Y", "%Y-%m"]):
    input_string = remove_quotes(input_string)
    for date_format in date_formats:
        try:
            datetime_object = datetime.strptime(input_string, date_format)
            date_str = datetime_object.strftime("%Y-%m-%d")
            return date_str
        except ValueError:
            continue
    return False


def remove_quotes(input_string):
    if len(input_string) >= 2 and ((input_string[0] == "'" and input_string[-1] == "'") or
                                   (input_string[0] == "(" and input_string[-1] == ")")):
        return input_string[1:-1]
    else:
        return input_string


def find_calculates(text):
    sents = text.split(' ')
    # print('sents', sents)
    temp = []
    for sent in sents:
        if sent in OPERATOR:
            temp.append(sent)
            continue
        if is_numeric_using_regex(sent):
            temp.append("{:.2f}".format(float(sent)))
            continue
        if any(substring in sent for substring in [i + '(' for i in FUNCTION]):
            f = find_functions(sent)
            temp.append(f[0] + '(' + f[1] + ')')
            continue
        if '.' in sent:
            temp.append('.'.join(sent.split('.')[1:]))
            continue
        # match = re.search(r'\b(\w+)\.(\w+)\.(\w+)\b', sent)
        # if match:
        #     print(1)
        #     temp.append('.'.join([match.group(2), match.group(3)]))
        #     continue
        # match = re.search(r'(\w+)\.(\w+)', sent)
        # if match:
        #     print(2)
        #     temp.append(match.group(2))
        #     continue
        if sent not in OPERATOR and has_english_characters(sent) and '[' not in sent:
            temp.append('[MASK]')
            continue
        temp.append(sent)
    calculate = ' '.join(temp)
    return calculate


def find_functions(text):
    # print(text)
    for i in FUNCTION:
        if i + '(' in text:
            # print(text)
            pattern = r'\b(?:' + i + r')\s*\((.*?)\)'
            match = re.search(pattern, text)
            if match:
                content = match.group(1)
                if is_numeric_using_regex(content):
                    return (i, content)
                if is_datetime_format(content):
                    return (i, is_datetime_format(content))
                content = '.'.join(content.split('.')[1:])
                # print('content', content)
                return (i, content)
            else:
                # print(text)
                return (i, text.replace(i, '')[1:-1])
    return ''


def find_tag_attrs(text):
    tag_attrs = '.'.join(text.split('.')[1:])
    return tag_attrs


def read_dict(dict):
    for key, value in dict.items():
        print(key, ': ', value)


def find_orderby(text):
    # print(text)
    word_list = text.split(' ')
    order_index = word_list.index('order')
    try:
        order_attr = '.'.join(word_list[order_index - 1].split('.')[1:])
        # print('order_attr', order_attr)
        # order_attr_var1 = word_list[order_index - 1]
        # order_attr_var2 = word_list[order_index + 2]
        try:
            order_sort = word_list[order_index + 3]
        except:
            order_sort = 'asc'
        orderby = (order_attr, order_sort)
        return orderby
    except:
        try:
            order_sort = word_list[order_index + 2]
        except:
            order_sort = 'asc'
        orderby = (word_list[order_index + 2], order_sort)
        return orderby


def special_sort(l):
    return sorted(l, key=lambda x: x[0] if isinstance(x, tuple) else x)


if __name__ == '__main__':
    content = '2023'
    s = is_datetime_format(content)
    print(s)



# -*- coding: utf-8 -*-
# @File    :   test.py
# @Time    :   2023/03/09 16:07:55
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
'''
import abc
import torch
from torch.utils.data import DataLoader
import datasets
from ..data import load_json, save_json

class GenerationPipeline():
    """ 传入model 和 tokenizer
        自行实现save_fn保存结果
        __call__ 进行预测和生成
    """
    def __init__(self, model, tokenizer) -> None:
        
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')  
        self.model = model.to(self.device)
        self.tokenizer = tokenizer

    @abc.abstractmethod
    def save_fn(self, ):

        pass 

    @abc.abstractmethod
    def __call__(self, data, save_fn=None):
        

        return 


class TerminalTestor:
    """
        方便进行逐条测试
        1. 继承此类，重写/传入 forward_fn 和 __call__
    """
    def __init__(self) -> None:
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') 

    def input_fn(self):
        s = input(f"Type in your input data:\n")
        return s 
    

    def forward(self, model, td, tokenizer, generation_config=None):
        """
            model: torch model
            td: tokenized results from tokenizer
            tokenizer: transformers tokenizer
            generation_config: transformers generation config, transformers version >= 4.28
        """
        if generation_config is None:
            pred_ids = model.generate(
                input_ids=td['input_ids'] , 
                attention_mask=td['attention_mask'], 
                max_length=500, num_beams=1
            )
        else:
            pred_ids = model.generate(
                input_ids=td['input_ids'] , 
                attention_mask=td['attention_mask'], 
                generation_config=generation_config 
            )
        texts = tokenizer.batch_decode(pred_ids, clean_up_tokenization_spaces=True, skip_special_tokens=True)
        return texts


    def __call__(self, model, tokenizer, forward_fn):
        """调用方法

        Args:
            model: torch model
            tokenizer: transformers tokenizer
            forward_fn: 自行编写 or 使用 self.forward
        """
        model = model.to(self.device)
        with torch.no_grad():
            while True:
                inputs = self.input_fn()
                td = tokenizer(inputs, return_tensors='pt').to(self.device)
                output = forward_fn(model, td, tokenizer)
                print(output)
                print("="*80)


def terminal_test(config, module):

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    # model = Wrapper.load_from_checkpoint(checkpoint_path=config.preds.ckpt_path, config=config).to(device)
    model = module(config).to(device)
    
    with torch.no_grad():
        while True:
            q = input("input your question:\n")
            q = model.tokenizer.bos_token + "Here is a question: " + q + " The CQL for this question is: "
            td = model.tokenizer(q, return_tensors='pt').to(device)
            res = model.model.generate(
                input_ids=td['input_ids'].to(device), attention_mask=td['attention_mask'].to(device), 
                max_length=500, num_beams=1, pad_token_id=50256)

            texts = model.tokenizer.batch_decode(res, clean_up_tokenization_spaces=True, skip_special_tokens=True)
            print(texts)

if __name__ == '__main__':
    pass 




import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from itertools import filterfalse


class CircleLoss(nn.Module):
    """ https://arxiv.org/abs/2002.10857v2  Equation.4 """
    def __init__(self, optim_pos=1, optim_neg=0, temperature=10):
        """ 做完softmax之后分类 最优情况下 optim_p,  optim_n = 1, 0 
            cosim 得分的话  optim_p,  optim_n = 1, -1
        """
        super().__init__()
        self.optim_pos = optim_pos
        self.optim_neg = optim_neg
        self.temperature = temperature

    def forward(self, pred, gold):
        """ replacing ce, pred should be softmaxed """
        # pred = safesoftmax(pred)
        pred = torch.softmax(pred, dim=-1)
        bsz, num_classes = pred.size()
        pos_ones = F.one_hot(gold, num_classes)
        neg_ones = torch.ones_like(pos_ones) - pos_ones
        
        alpha_p = torch.clamp_min(self.optim_pos - pred, min=0)
        alpha_n = torch.clamp_min(pred - self.optim_neg, min=0)
        p_dists = torch.exp(-pred * self.temperature * alpha_p) * pos_ones
        n_dists = torch.exp(pred * self.temperature * alpha_n ) * neg_ones

        p_dists = torch.sum(p_dists, 1, True)
        n_dists = torch.sum(n_dists, 1, True)
        loss = torch.log(1+ (n_dists)*(p_dists))
        loss = loss.mean()
        return loss

class FocalLoss(nn.Module):
    def __init__(self, gamma=1.6, alpha=0.4, size_average=True):
        """ alpha 是乘到负样本的loss上的权重，正样本是(1-alpha); None的话，两者相同，或者设为0.5?
            取该类别样本数目比例的倒数？
        """
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        if isinstance(alpha, (float,int)): self.alpha = torch.Tensor([alpha,1-alpha])
        if isinstance(alpha, list): self.alpha = torch.Tensor(alpha)
        self.size_average = size_average

    def forward(self, input, target):
        if input.dim() > 2:
            input = input.view(input.size(0), input.size(1),-1)
            input = input.transpose(1, 2)
            input = input.contiguous().view(-1, input.size(2))
        target = target.view(-1, 1)

        logpt = F.log_softmax(input, dim=-1)
        logpt = logpt.gather(1, target)
        logpt = logpt.view(-1)
        pt = logpt.data.exp()

        if self.alpha is not None:
            if self.alpha.type() != input.data.type():
                self.alpha = self.alpha.type_as(input.data)
            at = self.alpha.gather(0, target.data.view(-1))
            logpt = logpt * at

        loss = -1 * (1-pt)**self.gamma * logpt
        if self.size_average: 
            return loss.mean()
        else: 
            return loss.sum()


class DiceLoss(nn.Module):
    """ Taken from  https://github.com/lyakaap/pytorch-template """
    def __init__(self, smooth=1.0, eps=1e-7):
        super(DiceLoss, self).__init__()
        self.smooth = smooth
        self.eps = eps

    def forward(self, output, target):
        output = torch.sigmoid(output)

        if torch.sum(target) == 0:
            output = 1.0 - output
            target = 1.0 - target

        return 1.0 - (2 * torch.sum(output * target) + self.smooth) / (
                torch.sum(output) + torch.sum(target) + self.smooth + self.eps)


class FocalLoss(nn.Module):
    """ Taken from  https://github.com/lyakaap/pytorch-template """
    def __init__(self, gamma=2, eps=1e-7):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.eps = eps

    def forward(self, logit, target):
        prob = torch.sigmoid(logit)
        prob = prob.clamp(self.eps, 1. - self.eps)

        loss = -1 * target * torch.log(prob)
        loss = loss * (1 - logit) ** self.gamma

        return loss.sum()


# -*- coding: utf-8 -*-
# @File    :   qwen_chat_tokenizer.py
# @Time    :   2024/06/26 16:03:04
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    自定义 Qwen2Tokenizer 对于中文词汇分词时分到 char 级别
    原理: bpe 时遇到中文字符停止合并
'''


import regex as re
from transformers import Qwen2Tokenizer



# Copied from transformers.models.gpt2.tokenization_gpt2.get_pairs
def get_pairs(word):
    """
    Return set of symbol pairs in a word.

    Word is represented as tuple of symbols (symbols being variable-length strings).
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs

class QwenCharTokenizer(Qwen2Tokenizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.chinese_char_range = (0x4E00, 0x9FFF)
        self.chinese_chars_ids = self._get_chinese_char_ids()

    def _get_chinese_char_ids(self):
        """ 找出 qwen 词表中所有的中文单字的 id"""
        R = set()
        for i in range(0, len(self)):
            text = self.decode([i])
            # if len(text) != 1:
                # print(i, text)
            if re.match(r'\p{Han}', text):
                if len(text) == 1:
                    R.add(i)
        return R

    
    def is_chinese_char(self, token):
        """判断 token 是否是中文字符"""
        token_id = self._convert_token_to_id(token)
        if token_id in self.chinese_chars_ids:
            return True 
        else:
            return False

    
    # Copied from transformers.models.gpt2.tokenization_gpt2.GPT2Tokenizer.bpe
    def bpe(self, token):
        if token in self.cache:
            return self.cache[token]
        word = tuple(token)
        pairs = get_pairs(word)

        if not pairs:
            return token

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram

            # Check if we are combining Chinese characters and stop if they are
            if self.is_chinese_char(first) or self.is_chinese_char(second):
                break

            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                except ValueError:
                    new_word.extend(word[i:])
                    break
                else:
                    new_word.extend(word[i:j])
                    i = j

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = " ".join(word)
        self.cache[token] = word
        return word

    def _tokenize(self, text):
        """Tokenize a string."""
        bpe_tokens = []
        for token in re.findall(self.pat, text):
            token = "".join(
                self.byte_encoder[b] for b in token.encode("utf-8")
            )  
            # here are byte-level ascii characters 0-255

            # Maps all our bytes to unicode strings, avoiding control tokens of the BPE (spaces in our case)
            bpe_tokens.extend(bpe_token for bpe_token in self.bpe(token).split(" ")) 
            # 合并 byte 直到 byte pair不在词表中，bpe_tokens 就是合并后剩下的字符
        return bpe_tokens
    

if __name__ == '__main__':
    
    PATH = "/sshfs/pretrains/Qwen/Qwen2-0.5B-Instruct/"
    # 使用自定义的 CharTokenizer
    char_tokenizer = QwenCharTokenizer.from_pretrained(PATH)

    # 测试
    test_sentence = "醉春风"
    tokens = char_tokenizer.tokenize(test_sentence)
    token_ids = char_tokenizer.convert_tokens_to_ids(tokens)

    print("Tokens:", tokens)
    print("Token IDs:", token_ids)
    print()



import rdflib
from rdflib import URIRef, BNode, Literal, XSD
from rdflib.plugins.stores import sparqlstore
from itertools import chain
import os, sys 
try:
    from q_snippets.data import load_json, save_json
except:
    from qdls.data import load_json, save_json
from qdls.gql.sparql.utils.syntax import syntax_check
from qdls.utils import print_string
from multiprocessing import Pool

# FILE_DIR = os.path.dirname(os.path.abspath(__file__))
# PROJ_DIR = FILE_DIR[:FILE_DIR.index('src')]
# PROJ_DIR = "/home/qing/workspace/archive/ACL23"
# sys.path.append(PROJ_DIR)
from queue import Queue

import json 
import re 
from datetime import date
from tqdm import tqdm

class ValueClass():
    def __init__(self, type, value, unit=None):
        """
        When type is
            - string, value is a str
            - quantity, value is a number and unit is required
            - year, value is a int
            - date, value is a date object
        """
        self.type = type
        self.value = value
        self.unit = unit

    def isTime(self):
        return self.type in {'year', 'date'}

    def can_compare(self, other):
        if self.type == 'string':
            return other.type == 'string'
        elif self.type == 'quantity':
            # NOTE: for two quantity, they can compare only when they have the same unit
            return other.type == 'quantity' and other.unit == self.unit
        else:
            # year can compare with date
            return other.type == 'year' or other.type == 'date'

    def contains(self, other):
        """
        check whether self contains other, which is different from __eq__ and the result is asymmetric
        used for conditions like whether 2001-01-01 in 2001, or whether 2001 in 2001-01-01
        """
        if self.type == 'year': # year can contain year and date
            other_value = other.value if other.type == 'year' else other.value.year
            return self.value == other_value
        elif self.type == 'date': # date can only contain date
            return other.type == 'date' and self.value == other.value
        else:
            raise Exception('not supported type: %s' % self.type)


    def __eq__(self, other):
        """
        2001 and 2001-01-01 is not equal
        """
        assert self.can_compare(other)
        return self.type == other.type and self.value == other.value

    def __lt__(self, other):
        """
        Comparison between a year and a date will convert them both to year
        """
        assert self.can_compare(other)
        if self.type == 'string':
            raise Exception('try to compare two string')
        elif self.type == 'quantity':
            return self.value < other.value
        elif self.type == 'year':
            other_value = other.value if other.type == 'year' else other.value.year
            return self.value < other_value
        elif self.type == 'date':
            if other.type == 'year':
                return self.value.year < other.value
            else:
                return self.value < other.value

    def __gt__(self, other):
        assert self.can_compare(other)
        if self.type == 'string':
            raise Exception('try to compare two string')
        elif self.type == 'quantity':
            return self.value > other.value
        elif self.type == 'year':
            other_value = other.value if other.type == 'year' else other.value.year
            return self.value > other_value
        elif self.type == 'date':
            if other.type == 'year':
                return self.value.year > other.value
            else:
                return self.value > other.value

    def __str__(self):
        if self.type == 'string':
            return self.value
        elif self.type == 'quantity':
            if self.value - int(self.value) < 1e-5:
                v = int(self.value)
            else:
                v = self.value
            return '{} {}'.format(v, self.unit) if self.unit != '1' else str(v)
        elif self.type == 'year':
            return str(self.value)
        elif self.type == 'date':
            return self.value.isoformat()


class DataForSPARQL(object):
    def __init__(self, kb_path):
        kb = json.load(open(kb_path))
        self.concepts = kb['concepts']
        self.entities = kb['entities']

        # replace adjacent space and tab in name, which may cause errors when building sparql query
        for con_id, con_info in self.concepts.items():
            con_info['name'] = ' '.join(con_info['name'].split())
        for ent_id, ent_info in self.entities.items():
            ent_info['name'] = ' '.join(ent_info['name'].split())

        # get all attribute keys and predicates
        self.attribute_keys = set()
        self.predicates = set()
        self.key_type = {}
        for ent_id, ent_info in self.entities.items():
            for attr_info in ent_info['attributes']:
                self.attribute_keys.add(attr_info['key'])
                self.key_type[attr_info['key']] = attr_info['value']['type']
                for qk in attr_info['qualifiers']:
                    self.attribute_keys.add(qk)
                    for qv in attr_info['qualifiers'][qk]:
                        self.key_type[qk] = qv['type']
        for ent_id, ent_info in self.entities.items():
            for rel_info in ent_info['relations']:
                self.predicates.add(rel_info['predicate'])
                for qk in rel_info['qualifiers']:
                    self.attribute_keys.add(qk)
                    for qv in rel_info['qualifiers'][qk]:
                        self.key_type[qk] = qv['type']
        self.attribute_keys = list(self.attribute_keys)
        self.predicates = list(self.predicates)
        # Note: key_type is one of string/quantity/date, but date means the key may have values of type year
        self.key_type = { k:v if v!='year' else 'date' for k,v in self.key_type.items() }

        # parse values into ValueClass object
        for ent_id, ent_info in self.entities.items():
            for attr_info in ent_info['attributes']:
                attr_info['value'] = self._parse_value(attr_info['value'])
                for qk, qvs in attr_info['qualifiers'].items():
                    attr_info['qualifiers'][qk] = [self._parse_value(qv) for qv in qvs]
        for ent_id, ent_info in self.entities.items():
            for rel_info in ent_info['relations']:
                for qk, qvs in rel_info['qualifiers'].items():
                    rel_info['qualifiers'][qk] = [self._parse_value(qv) for qv in qvs]

    def _parse_value(self, value):
        if value['type'] == 'date':
            x = value['value']
            p1, p2 = x.find('/'), x.rfind('/')
            y, m, d = int(x[:p1]), int(x[p1+1:p2]), int(x[p2+1:])
            result = ValueClass('date', date(y, m, d))
        elif value['type'] == 'year':
            result = ValueClass('year', value['value'])
        elif value['type'] == 'string':
            result = ValueClass('string', value['value'])
        elif value['type'] == 'quantity':
            result = ValueClass('quantity', value['value'], value['unit'])
        else:
            raise Exception('unsupport value type')
        return result

    def get_direct_concepts(self, ent_id):
        """
        return the direct concept id of given entity/concept
        """
        if ent_id in self.entities:
            return self.entities[ent_id]['instanceOf']
        elif ent_id in self.concepts:
            return self.concepts[ent_id]['instanceOf']
        else:
            raise Exception('unknown id')

    def get_all_concepts(self, ent_id):
        """
        return a concept id list
        """
        ancestors = []
        q = Queue()
        for c in self.get_direct_concepts(ent_id):
            q.put(c)
        while not q.empty():
            con_id = q.get()
            ancestors.append(con_id)
            for c in self.concepts[con_id]['instanceOf']:
                q.put(c)

        return ancestors

    def get_name(self, ent_id):
        if ent_id in self.entities:
            return self.entities[ent_id]['name']
        elif ent_id in self.concepts:
            return self.concepts[ent_id]['name']
        else:
            return None

    def is_concept(self, ent_id):
        return ent_id in self.concepts

    def get_attribute_facts(self, ent_id, key=None, unit=None):
        if key:
            facts = []
            for attr_info in self.entities[ent_id]['attributes']:
                if attr_info['key'] == key:
                    if unit:
                        if attr_info['value'].unit == unit:
                            facts.append(attr_info)
                    else:
                        facts.append(attr_info)
        else:
            facts = self.entities[ent_id]['attributes']
        facts = [(f['key'], f['value'], f['qualifiers']) for f in facts]
        return facts

    def get_relation_facts(self, ent_id):
        facts = self.entities[ent_id]['relations']
        facts = [(f['predicate'], f['object'], f['direction'], f['qualifiers']) for f in facts]
        return facts



# kb = DataForSPARQL(os.path.join(PROJ_DIR, 'data/dataset/kb.json'))
# kb = DataForSPARQL("/Users/qing/Downloads/kb.json")

def legal(s):
    # convert predicate and attribute keys to legal format
    return s.replace(' ', '_')

class SparqlEngine():
    gs1 = None
    PRED_INSTANCE = 'pred:instance_of'
    PRED_NAME = 'pred:name'

    PRED_VALUE = 'pred:value'       # link packed value node to its literal value
    PRED_UNIT = 'pred:unit'         # link packed value node to its unit

    PRED_YEAR = 'pred:year'         # link packed value node to its year value, which is an integer
    PRED_DATE = 'pred:date'         # link packed value node to its date value, which is a date

    PRED_FACT_H = 'pred:fact_h'     # link qualifier node to its head
    PRED_FACT_R = 'pred:fact_r'
    PRED_FACT_T = 'pred:fact_t'

    SPECIAL_PREDICATES = (PRED_INSTANCE, PRED_NAME, PRED_VALUE, PRED_UNIT, PRED_YEAR, PRED_DATE, PRED_FACT_H, PRED_FACT_R, PRED_FACT_T)
    def __init__(self, data, ttl_file=''):
        self.nodes = nodes = {}
        for i in chain(data.concepts, data.entities):
            nodes[i] = URIRef(i)
        for p in chain(data.predicates, data.attribute_keys, SparqlEngine.SPECIAL_PREDICATES):
            nodes[p] = URIRef(legal(p))
        
        self.graph = graph = rdflib.Graph()

        for i in chain(data.concepts, data.entities):
            name = data.get_name(i)
            graph.add((nodes[i], nodes[SparqlEngine.PRED_NAME], Literal(name)))

        for ent_id in tqdm(data.entities, desc='Establishing rdf graph'):
            for con_id in data.get_all_concepts(ent_id):
                graph.add((nodes[ent_id], nodes[SparqlEngine.PRED_INSTANCE], nodes[con_id]))
            for (k, v, qualifiers) in data.get_attribute_facts(ent_id):
                h, r = nodes[ent_id], nodes[k]
                t = self._get_value_node(v)
                graph.add((h, r, t))
                fact_node = self._new_fact_node(h, r, t)

                for qk, qvs in qualifiers.items():
                    for qv in qvs:
                        h, r = fact_node, nodes[qk]
                        t = self._get_value_node(qv)
                        if len(list(graph[t])) == 0:
                            print(t)
                        graph.add((h, r, t))

            for (pred, obj_id, direction, qualifiers) in data.get_relation_facts(ent_id):
                if direction == 'backward':
                    if data.is_concept(obj_id):
                        h, r, t = nodes[obj_id], nodes[pred], nodes[ent_id]
                    else:
                        continue
                else:
                    h, r, t = nodes[ent_id], nodes[pred], nodes[obj_id]
                graph.add((h, r, t))
                fact_node = self._new_fact_node(h, r, t)
                for qk, qvs in qualifiers.items():
                    for qv in qvs:
                        h, r = fact_node, nodes[qk]
                        t = self._get_value_node(qv)
                        graph.add((h, r, t))

        if ttl_file:
            print('Save graph to {}'.format(ttl_file))
            graph.serialize(ttl_file, format='turtle')


    def _get_value_node(self, v):
        # we use a URIRef node, because we need its reference in query results, which is not supported by BNode
        if v.type == 'string':
            node = BNode()
            self.graph.add((node, self.nodes[SparqlEngine.PRED_VALUE], Literal(v.value)))
            return node
        elif v.type == 'quantity': 
            # we use a node to pack value and unit
            node = BNode()
            self.graph.add((node, self.nodes[SparqlEngine.PRED_VALUE], Literal(v.value, datatype=XSD.double)))
            self.graph.add((node, self.nodes[SparqlEngine.PRED_UNIT], Literal(v.unit)))
            return node
        elif v.type == 'year':
            node = BNode()
            self.graph.add((node, self.nodes[SparqlEngine.PRED_YEAR], Literal(v.value)))
            return node
        elif v.type == 'date':
            # use a node to pack year and date
            node = BNode()
            self.graph.add((node, self.nodes[SparqlEngine.PRED_YEAR], Literal(v.value.year)))
            self.graph.add((node, self.nodes[SparqlEngine.PRED_DATE], Literal(v.value, datatype=XSD.date)))
            return node

    def _new_fact_node(self, h, r, t):
        node = BNode()
        self.graph.add((node, self.nodes[SparqlEngine.PRED_FACT_H], h))
        self.graph.add((node, self.nodes[SparqlEngine.PRED_FACT_R], r))
        self.graph.add((node, self.nodes[SparqlEngine.PRED_FACT_T], t))
        return node



def query_virtuoso(q, virtuoso_address, virtuoso_graph_uri):
    endpoint = virtuoso_address
    store=sparqlstore.SPARQLUpdateStore(endpoint)
    gs = rdflib.ConjunctiveGraph(store)
    gs.open((endpoint, endpoint))
    gs1 = gs.get_context(rdflib.URIRef(virtuoso_graph_uri))
    res = gs1.query(q)
    return res

def get_sparql_answer(sparql, data, config):
    """
    data: DataForSPARQL object, we need the key_type
    """
    try:
        # infer the parse_type based on sparql
        if sparql.startswith('SELECT DISTINCT ?e') or sparql.startswith('SELECT ?e'):
            parse_type = 'name'
        elif sparql.startswith('SELECT (COUNT(DISTINCT ?e)'):
            parse_type = 'count'
        elif sparql.startswith('SELECT DISTINCT ?p '):
            parse_type = 'pred'
        elif sparql.startswith('ASK'):
            parse_type = 'bool'
        else:
            tokens = sparql.split()
            tgt = tokens[2]
            for i in range(len(tokens)-1, 1, -1):
                if tokens[i]=='.' and tokens[i-1]==tgt:      #  key tgt .
                    key = tokens[i-2]
                    break
            key = key[1:-1].replace('_', ' ')
            t = data.key_type[key]
            parse_type = 'attr_{}'.format(t)

        parsed_answer = None

        res = query_virtuoso(sparql, virtuoso_address=config.virtuoso_address, virtuoso_graph_uri=config.virtuoso_graph_uri)

        if res.vars:
            res = [[binding[v] for v in res.vars] for binding in res.bindings]
            if len(res) != 1:
                return None
        else:
            res = res.askAnswer
            assert parse_type == 'bool'
        
        if parse_type == 'name':
            node = res[0][0]
            sp = 'SELECT DISTINCT ?v WHERE {{ <{}> <{}> ?v .  }}'.format(node, SparqlEngine.PRED_NAME)
            res = query_virtuoso(sp, virtuoso_address=config.virtuoso_address, virtuoso_graph_uri=config.virtuoso_graph_uri)
            res = [[binding[v] for v in res.vars] for binding in res.bindings]
            name = res[0][0].value
            parsed_answer = name
        elif parse_type == 'count':
            count = res[0][0].value
            parsed_answer = str(count)
        elif parse_type.startswith('attr_'):
            node = res[0][0]
            v_type = parse_type.split('_')[1]
            unit = None
            if v_type == 'string':
                sp = 'SELECT DISTINCT ?v WHERE {{ <{}> <{}> ?v .  }}'.format(node, SparqlEngine.PRED_VALUE)
            elif v_type == 'quantity':
                # Note: For those large number, ?v is truncated by virtuoso (e.g., 14756087 to 1.47561e+07)
                # To obtain the accurate ?v, we need to cast it to str
                sp = 'SELECT DISTINCT ?v,?u,(str(?v) as ?sv) WHERE {{ <{}> <{}> ?v ; <{}> ?u .  }}'.format(node, SparqlEngine.PRED_VALUE, SparqlEngine.PRED_UNIT)
            elif v_type == 'year':
                sp = 'SELECT DISTINCT ?v WHERE {{ <{}> <{}> ?v .  }}'.format(node, SparqlEngine.PRED_YEAR)
            elif v_type == 'date':
                sp = 'SELECT DISTINCT ?v WHERE {{ <{}> <{}> ?v .  }}'.format(node, SparqlEngine.PRED_DATE)
            else:
                raise Exception('unsupported parse type')
            res = query_virtuoso(sp, virtuoso_address=config.virtuoso_address, virtuoso_graph_uri=config.virtuoso_graph_uri)
            res = [[binding[v] for v in res.vars] for binding in res.bindings]
            # if there is no specific date, then convert the type to year
            if len(res)==0 and v_type == 'date':
                v_type = 'year'
                sp = 'SELECT DISTINCT ?v WHERE {{ <{}> <{}> ?v .  }}'.format(node, SparqlEngine.PRED_YEAR)
                res = query_virtuoso(sp, virtuoso_address=config.virtuoso_address, virtuoso_graph_uri=config.virtuoso_graph_uri)
                res = [[binding[v] for v in res.vars] for binding in res.bindings]
            if v_type == 'quantity':
                value = float(res[0][2].value)
                unit = res[0][1].value
            else:
                value = res[0][0].value
            value = ValueClass(v_type, value, unit)
            parsed_answer = str(value)
        elif parse_type == 'bool':
            parsed_answer = 'yes' if res else 'no'
        elif parse_type == 'pred':
            parsed_answer = str(res[0][0])
            parsed_answer = parsed_answer.replace('_', ' ')
        # return parsed_answer
    except Exception as e:
        print(e)
        parsed_answer = None 
    # print(parsed_answer)
    return parsed_answer

def whether_equal(answer, pred):
    """
    check whether the two arguments are equal as attribute value
    answer type should be str 
    """
    def truncate_float(x):
        # convert answer from '100.0 meters' to '100 meters'
        try:
            v, *u = x.split()
            v = float(v)
            if v - int(v) < 1e-5:
                v = int(v)
            if len(u) == 0:
                x = str(v)
            else:
                x = '{} {}'.format(str(v), ' '.join(u))
        except:
            pass
        return x

    def equal_as_date(x, y):
        # check whether x and y are equal as type of date or year
        try:
            x_split = x.split('-')
            y_split = y.split('-')
            if len(x_split) == 3:
                x = date(int(x_split[0]), int(x_split[1]), int(x_split[2]))
            else:
                x = int(x)
            if len(y_split) == 3:
                y = date(int(y_split[0]), int(y_split[1]), int(y_split[2]))
            else:
                y = int(y)
            if isinstance(x, date) and isinstance(y, date):
                return x == y
            else:
                x = x.year if isinstance(x, date) else x
                y = y.year if isinstance(y, date) else y
                return x == y
        except:
            return False

    answer = truncate_float(answer)
    pred = truncate_float(pred)
    if equal_as_date(answer, pred):
        return True
    else:
        return answer == pred


def post_process(text):
    """
    针对BART生成结果的预处理
    字符串提取出来，其余部分是chunks， 需要增加?和.前的空格， 最后再组装起来即可
    !!!!!!!!!  只需要处理生成的， trainset里的就不用处理了
    """
    pattern = re.compile(r'".*?"')    # 匹配字符串？
    nes = []
    for item in pattern.finditer(text):
        nes.append((item.group(), item.span()))
    pos = [0]
    for name, span in nes:
        pos += [span[0], span[1]]
    pos.append(len(text))
    assert len(pos) % 2 == 0
    assert len(pos) / 2 == len(nes) + 1
    chunks = [text[pos[i]: pos[i+1]] for i in range(0, len(pos), 2)]
    for i in range(len(chunks)):
        chunks[i] = chunks[i].replace('?', ' ?').replace('.', ' .')           # 增加空格
    bingo = ''
    for i in range(len(chunks) - 1):
        bingo += chunks[i] + nes[i][0]
    bingo += chunks[-1]
    return bingo

def pseudo_post_process(text):
    return text 

def exec_sparql(sparql, answer, config, kb, process_fn=pseudo_post_process):
    sparql = process_fn(sparql)
    # print("==>", sparql)
    pred_answer = get_sparql_answer(sparql, kb, config)
    is_match = whether_equal(answer, pred_answer)
    # print(sparql)
    # print(is_match, pred_answer, answer)
    return is_match, pred_answer


def sparql_exec_acc(path, key='pred', nproc=1, resave=False, config=None, kb=None, process_fn=pseudo_post_process):
    """ path: 可以是字符串（读取文件）也可以是读取后的对象list 
        key: 每一个sample字典中, 预测的sparql语句的key
        nproc: 并行处理的进程数
        resave: bool or string  是否重新保存结果, 保存在原目录, 文件名前加上exec_
        config: 配置文件, 主要是virtuoso的 address 和 graph uri
        kb: 知识库
        process_fn: 对sparql进行处理的函数
    """
    data = load_json(path) if type(path) is str else path 
    Results = []
    if nproc > 1:
        with Pool(nproc) as pool:
            R = {}
            for sample in data:
                assert key in sample, f"key `{key}` not in provided data, sample.keys()={sample.keys()}"
                sparql = sample[key]
                answer = sample['answer']
                future = pool.apply_async(exec_sparql, (sparql, answer, config, kb, process_fn))
                R[future] = sample
            
            for future in tqdm(R):
                try:
                    res = future.get(timeout=3)
                except:
                    res = False, 'timeout'  
                Results.append(res)
    else:
        print_string("sequential processing")
        for sample in tqdm(data):
            assert key in sample, f"key `{key}` not in provided data, sample.keys()={sample.keys()}"
            sparql = sample[key]
            answer = sample['answer']
            Results.append(exec_sparql(sparql, answer, config, kb, process_fn))

    cnt = len(data)
    correct, wrong = 0, 0
    for is_match, pred in Results:
        if is_match:
            correct += 1
        else:
            wrong += 1
 

    acc = 100 *  correct/ cnt
    print(f"{correct} of {cnt} is correct, {acc:.3f};  {wrong} wrong")

    if resave:
        if type(resave) is not str:
            assert type(path) is str, "resave=True, path should be a string, or set resave=filepath"
        else:
            path = resave

        for sample, exec_res in zip(data, Results):
            sample['exec'] = exec_res
        
        new_filename = "exec_" + os.path.basename(path)
        new_path = os.path.join(os.path.dirname(path), new_filename)
        save_json(data, new_path)

    return acc, Results



if __name__ == '__main__':
    pass 


import os
try:
    import torch
except:
    pass
import json
import yaml
import math
import time
import hashlib
from collections import defaultdict
import numpy as np
import pandas as pd

from tqdm import tqdm
import pickle

def timeit(f):
    """ 函数装饰器，定义函数时调用 """
    def timed(*args, **kw):
        ts = time.time()
        result = f(*args, **kw)
        te = time.time()

        print('func:%r args:[%r, %r] took: %2.4f sec' % (f.__name__, args, kw, te-ts))
        return result

    return timed

def save_json(R, path, **kwargs):
    """ Obj, path """
    with open(path, 'w', encoding='utf8') as f:
        json.dump(R, f, indent=2, ensure_ascii=False, **kwargs)
    print(f"{path} saved with {len(R)} samples!")

def load_json(path):
    with open(path, 'r', encoding='utf8') as f:
        obj = json.load(f)
    print(f"{path} loaded with {len(obj)} samples!")
    return obj

def load_jsonl(path):
    R = []
    with open(path, 'r', encoding='utf8') as f:
        for line in f:
            R.append(json.loads(line))
    print(f"{path} loaded with {len(R)} samples!")
    return R


def load_yaml(path):
    with open(path) as fin:
        return yaml.safe_load(fin)


def save_pickle(R, path):
    """ save some objects that json not support, like DateTime """
    with open(path, 'wb') as f:
        pickle.dump(R, f)
    print(f"{path} saved with {len(R)} samples!")
    
def load_pickle(path):
    with open(path, 'rb') as f:
        obj = pickle.load(f)
    print(f"{path} loaded with {len(obj)} samples!")
    return obj


class BaseData:    
    def cpu(self):
        for k, v in self.__dict__.items():
            if type(v) is torch.Tensor:
                setattr(self, k, v.cpu())
        return self
    
    def to(self, device):
        for k,v in self.__dict__.items():
            if type(v) is torch.Tensor:
                setattr(self, k, v.to(device))
        return self
    
    def __getitem__(self,index):
        return self.__dict__[index]
    
    def todict(self):
        return self.__dict__
    
    def tolist(self):
        return [ v for k,v in self.__dict__.items()]

    @property
    def size_info(self):
        size = { 
            k: v.size() if type(v) is torch.Tensor else 
            (len(v) if v is not None else None)
            for k,v in self.__dict__.items()
        }
        return size
    


def span_decode(start_logits, end_logits, cls_logits=None, max_a_len=512, samples=None, offset_mappings=None, use_cls=True, no_answer=""):
    """

    Args:
        start_logits  (torch.Tensor) :  (bsz,seqlen)
        end_logits (torch.Tensor) :   (bsz,seqlen)
        cls_logits (torch.Tensor) :  (bsz, num_classes)
        max_a_len ( int ): 限制答案文本的最大长度
        samples (MRCSample ): 该条数据的所有信息
        offset_mappings ([type]): tokenizer返回的
        use_cls (bool, optional): 是否使用预测的有无答案的概率，False则一定会返回预测的span文本. Defaults to True.
        no_answer (str, optional): Squad和DuReader要求的无答案形式不同，. Defaults to "".

    Returns:
        Dict : {qid: pred_text, ...}
    """
    se_sum = end_logits[:,None,:] + start_logits[:,:,None]
    # 限制值的范围是s<e, 最大长度为 max_a_len        
    mask = torch.tril(torch.triu(torch.ones_like(se_sum), 0), max_a_len)   
    r = (mask * se_sum).masked_fill_((1-mask).bool(), float('-inf'))    # score值全是负的，导致0 > score，选出来s>e了
    start_max, end_max = r.max(2)[0].max(1)[1], r.max(1)[0].max(1)[1]
    answerable = cls_logits.argmax(-1) if cls_logits is not None else torch.zeros(start_logits.size(0),)
    R = {}
    for s, e, a, sample, mapping in zip(start_max, end_max, answerable, samples, offset_mappings):
        if a == 1 and use_cls:
            R[sample.qid] = no_answer
        else:
            s_char, e_char = mapping[s][0], mapping[e][-1]
            pred_text = sample.context[s_char:e_char]
            pred_text = no_answer if pred_text == "" else pred_text
            R[sample.qid] = pred_text
    return R


def nbest_span_decode(start_scores, end_scores, batch, max_a_len=512, nbest=5):
    # se_sum = end_scores[:,None,:] + start_scores[:,:,None]
    # r = torch.tril(torch.triu(torch.ones_like(se_sum), 0), max_a_len) * se_sum   # (0,0) is necessary !!!!!!!!!!!!
    se_sum = end_scores[:,None,:] + start_scores[:,:,None]
    # 限制值的范围是s<e, 最大长度为 max_a_len        
    mask = torch.tril(torch.triu(torch.ones_like(se_sum), 0), max_a_len)   
    r = (mask * se_sum).masked_fill_((1-mask).bool(), float('-inf'))

    pred_answers = defaultdict(dict)
    for mat, c, m, qid, a in zip(r, batch['context'], batch['offset_mapping'], batch['qids'], batch['gold']):
        pred_answers[qid]['gold'] = a
        pred_answers[qid]['nbest'] = []
        v, i = torch.topk(mat.flatten(), int(nbest))
        v, i = v.cpu(), i.cpu()
        for (s,e), p in zip(np.array(np.unravel_index(i.numpy(), mat.shape)).T, v):
            s_char, e_char =  m[s][0], m[e][-1]
            pred_text = c[s_char:e_char]
            pred_text = "no answer" if pred_text == "" else pred_text
            pred_answers[qid]['nbest'].append({"prob":f"{p.item():.3f}", "text": pred_text})
    return pred_answers

def gen_uid(string):
    """
    根据字符串生成对应的Md5码, 用与给数据生成ID
    Args:
        string (str): 需要是能标识该对象的独特字符串

    Returns:
        str: 十六进制字符串
    """
    return hashlib.md5(string.encode("utf8")).hexdigest()


def sequence_padding(inputs, length=None, force=False, padding=0, padding_side='right'):
    """Numpy函数，将序列padding到同一长度
    Args:
        inputs (list of lists): [description]
        length (int, optional): 指定最大长度. Defaults to None.
        force (bool, optional): 如果为True则一定Padding到length长度，否则padding到最长的list的长度. Defaults to False.
        padding (int, optional): Padding的值. Defaults to 0.
        padding_side: GPT2做预测的时候，padding to left

    Returns:
        np.array: padding后的序列
    """
    _length = max([len(x) for x in inputs])
    if length is None:
        length = _length
    else:
        length = min(length, _length) if not force else length

    outputs = np.array([
        np.concatenate(
            [x, [padding] * (length - len(x))] if padding_side == 'right' else [[padding] * (length - len(x)), x] 
        ) if len(x) < length else x[:length]
        for x in inputs
    ])

    return outputs 


def dict2list(td):
    """
        输入的是tokenizer一次处理多条数据的结果， 返回list, 其中每个dict是每条数据的结果
    Args:
        td (dict): {'input_ids': [[...],[...],...], 'attention_mask':...}

    Returns:
        [list]: [{'input_ids':[], 'attention_mask':[]},{},...]
    """
    df = pd.DataFrame.from_dict(td, orient='index').T
    return df.to_dict(orient='records')



class SegmentUtility():
    def __init__(self, seg_func) -> None:
        """
        Args:
            seg_func (function):  eg: lambda s: seg.cut(s) || jieba.lcut(s)
        """
        self.seg = seg_func

    def seg_start_end(self, sentence):
        """
            返回句子分词的结果，以及每个词对应的字符start end
        """
        words = self.seg(sentence)
        s,e = 0,0
        R = [[(s:=e, e:=s+len(word))] for word in words]
        return words, sum(R, [])

    @staticmethod
    def generate_seg_ids(tokenizer, input_ids, offset_mapping, se):
        """对于一个句子的input_ids， 根据该句子的中文分词结果，对于每个词的tokens给一个id
        从[CLS]是0开始，第一个词是1,...
        Args:
            input_ids (list): Tokenizer 返回的input_ids
            offset_mapping (list): Tokenizer 返回的 offset_mapping
            se (list): [(0,2),(2,6),...,(s,e)]

        Returns:
            list : [0,1,1,2,2,2,2,...]
        """
        R, ptr, word_ptr = [], 0, 0
        for token_id, (s,e) in zip(input_ids, offset_mapping):
            if (s,e) == (0,0) and token_id != tokenizer.pad_token_id:
                R.append(ptr)
                ptr += 1
            elif token_id == 0:  # padding
                R.append(ptr)
            elif se[word_ptr][0] <= s and e < se[word_ptr][1]:
                R.append(ptr)
            elif e == se[word_ptr][1]:
                R.append(ptr)
                word_ptr += 1
                ptr += 1
        assert len(R) == len(input_ids), f"{(len(R), R)},{(len(input_ids), input_ids)}"
        return R

    def prepare_inputs(self, tokenizer, first, second=None):
        """[summary]

        Args:
            tokenizer ([type]): [description]
            first (str): 第一个句子
            second (str, optional): 第二个句子 Defaults to None.

        Returns:
            [type]: [description]
        """
        words_1 = self.seg(first)
        s,e = 0,0
        se_1 = [[(s:=e, e:=s+len(word))] for word in words_1]
        
        if second is not None:
            words_2 = self.seg(second)
            s,e = 0,0
            se_2 = [[(s:=e, e:=s+len(word))] for word in words_2]
            
        td = tokenizer(first, second, return_offsets_mapping=True)
        td['se'] =  sum(se_1 if second is None else se_1 + se_2, [])
        return td

    def prepare_batch_inputs(self, tokenizer, first, second=None):
        """[summary]

        Args:
            tokenizer ([type]): [description]
            first (list):  字符串的列表
            second (list, optional): 字符串的list. Defaults to None.

        Returns:
            dict: 分词器分词后的结果，增加了words = [], se = []
        """
        td = tokenizer(first, second, return_offsets_mapping=True, padding=True)
        SE = []
        Seg = []
        if second is not None:
            for ids, mapping, a, b in zip(td.input_ids, td.offset_mapping, first, second):
                words_1 = self.seg(a)
                s,e = 0,0
                se_1 = [[(s:=e, e:=s+len(word))] for word in words_1]

                words_2 = self.seg(b)
                s,e = 0,0
                se_2 = [[(s:=e, e:=s+len(word))] for word in words_2]

                se = sum(se_1 + se_2, [])
                SE.append(se)
                Seg.append((words_1, words_2))
        else:
            for ids, mapping, a in zip(td.input_ids, td.offset_mapping, first):
                words_1 = self.seg(a)
                s,e = 0,0
                se_1 = [[(s:=e, e:=s+len(word))] for word in words_1]
                se = sum(se_1, [])
                SE.append(se)
                Seg.append(words_1)
        td.se = SE
        td.words = Seg
        return td

######################
#   IO  functions 


def recursive_file_find(root):
    """ return all file contained in dir 'root' and it's subdirs
    """
    R = []
    for file in os.listdir(root):
        abs_path = os.path.join(root, file)
        if os.path.isfile(abs_path):
            R.append(abs_path)
        else:
            R.extend(recursive_file_find(abs_path))
    R = sorted(R)
    return R


if __name__ == '__main__':
    R = recursive_file_find("..")
    print(f"{len(R)} files found! ")

    
   



import os 
import pandas as pd 
from typing import Any
from loguru import logger
from qdls.data import load_json, save_json
from qdls.utils import print_string
from collections import defaultdict
from tqdm import tqdm
from argparse import Namespace
 
import datasets 

from .base import BaseEvaluator
from .metric_fns import calc_metrics_per_sample

example_neo4j_config = Namespace(
    neo4j_uri="neo4j://$IP:$PORT", neo4j_user="neo4j", neo4j_passwd="kqa", timeout=3
)

example_virtuoso_config = Namespace(
    virtuoso_address = "http://127.0.0.1:28890/sparql",
    virtuoso_graph_uri = 'kqaspcy'
)


class KqaAutoEvaluator(BaseEvaluator):
    def __init__(self, file=None, lang='cypher', nproc=1, **kwargs) -> None:
        """
            file: json file path or loaded json data 
            lang: cypher or sparql, or other
            nproc: number of processes
            kwargs: 
                neo4j_config = Namespace(neo4j_uri="neo4j://map:28892", neo4j_user="neo4j", neo4j_passwd="neo4j", timeout=3)
                metrics_to_calc=['exact_set_match', 'rouge-L']
        """
        self.lang = lang
        self.nproc = nproc
        self.neo4j_config = kwargs.get('neo4j_config', "please set neo4j_config in kwargs")
        if kwargs.get('metrics_to_calc') is not None:
            self.metrics_to_calc = kwargs['metrics_to_calc']
        else:
            self.metrics_to_calc = ['bleu', 'executable', 'exact_match']
        self._manage_metrics_to_calc(kwargs)

        print_string(f"metrics to calc: {self.metrics_to_calc}")
        print_string(f"neo4j config: {self.neo4j_config}")

        if file is not None:
            self.raw_data = load_json(file) if type(file) is str else file 
            self.length2sids = self.length_to_sampleids()
            
            self.data_to_eval = [ self.normalize_sample(s, lang) for s in self.raw_data ]
            self.__pre_calc_metrics()
            self.id2sample = {sample['sample_id']:sample for sample in self.data_to_eval}


        if lang == 'sparql':
            print_string(f"to evaluate for sparql generation, calling obj.eval_sparql()")
            self.eval_sparql()

    def _manage_metrics_to_calc(self, kwargs):
        """  
            增加在 init 函数通过 kwargs 传入新的 metrics_to_calc
            lang 是 cypher 和 sparql 默认增加 is_correct
        """
        if self.lang in ['cypher', 'sparql']:
            self.metrics_to_calc.append("is_correct")  # 默认是要计算准确率的
        if "metrics_to_calc" in kwargs:
            for metric in kwargs['metrics_to_calc']:
                if metric not in self.metrics_to_calc:
                    self.metrics_to_calc.append(metric)   

    def eval_sparql(self):
        from .sparql_utils import exec_one_sample 
        from multiprocessing import Pool 
        with Pool(self.nproc) as p:
            R = list(tqdm(p.imap(exec_one_sample, self.data_to_eval), total=len(self.data_to_eval), desc=f"executing sparql with {self.nproc} processes"))
        for sample, (is_correct, info) in zip(self.data_to_eval, R):
            sample['is_correct'] = is_correct
            sample['exec_info'] = info
        

    @classmethod
    def from_processed(cls, file, lang='cypher'):
        data = load_json(file)
        obj = cls(lang=lang)
        obj.data_to_eval = data
        obj.id2sample = {sample['sample_id']:sample for sample in obj.data_to_eval}
        # obj.length2sids = obj.length_to_sampleids()  # no raw data
        return obj 

    def save_processed(self, target_path=None):
        if target_path is None:
            target_path = f"./eval_results_for_{self.lang}.json"
        save_json(self.data_to_eval, target_path)  

    def __pre_calc_metrics(self):
        """ 使用datasets库 并行计算 """
        ds = datasets.Dataset.from_list(self.data_to_eval)
        print(ds)
        
        if self.nproc > 1:
            ds = ds.map(
                calc_metrics_per_sample, num_proc=self.nproc, 
                fn_kwargs={
                    'ref_key':self.lang, 
                    'metrics':self.metrics_to_calc,
                    'neo4j_config': self.neo4j_config
                }
            )
            self.data_to_eval = ds.to_list()
        else:
            self.data_to_eval = [] 
            for s in tqdm(ds):
                self.data_to_eval.append(
                    calc_metrics_per_sample(s, self.lang, self.metrics_to_calc, self.neo4j_config)
                )
            
    def _get_sample_length(self, sample):
        return len(sample['program'])
    
    def length_to_sampleids(self):
        """ 按照program（KoPL）的长度，分组 sample_id
        2 452 0.051
        3 1890 0.266
        4 2414 0.541
        5 1072 0.662
        6 1438 0.826
        7 460 0.878
        8 742 0.963
        9 235 0.989
        10 7 0.990
        11 77 0.999
        12 6 1.000
        14 4 1.000
        """
        length2sids = defaultdict(list)
        for sample in self.raw_data:
            length2sids[self._get_sample_length(sample)].append(sample['sample_id'])
        length2sids = dict(sorted(length2sids.items(), key=lambda x:x[0]))
        return length2sids
    
    def calc_marco_metrics(self, ids=None):
        """ 根据预先计算好的metrics, 为ids中的样本计算平均值 """
        if ids is None:
            ids = self.id2sample.keys()
        L = len(ids)
        scores = defaultdict(list)
        for i in ids:
            sample = self.id2sample[i]    
            for m in self.metrics_to_calc:
                # if necessary, eg, some has negative values
                # scores[m].append(max(float(sample.get(m, 0)), 0))
                scores[m].append(float(sample.get(m, 0)))  
        return {k:sum(v)/L for k,v in scores.items()}
    
    @staticmethod
    def complexity_fn(length2sids):
        """ 划分困难的标准 """
        Easy, Medium, Hard = [], [], [] 
        for k,v in length2sids.items():
            if k <=3:
                Easy.extend(v)
            elif k <= 6:
                Medium.extend(v)
            else:
                Hard.extend(v)

        return Easy, Medium, Hard
    
    def detailed_length_metrics(self):
        """ 将问题按照长度分类，计算每个类别的问题的指标 """
        Easy, Medium, Hard = self.complexity_fn(self.length2sids)
        Easy_metrics = self.calc_marco_metrics(Easy)
        Medium_metrics = self.calc_marco_metrics(Medium)
        Hard_metrics = self.calc_marco_metrics(Hard)
        return Easy_metrics, Medium_metrics, Hard_metrics

    def exec_results(self):
        num = sum([x['exec_acc'] for k,x in self.id2sample.items()])
        acc = num /len(self.id2sample)
        print(f"exec acc: {acc}, {num} ")


import os
import re
import collections
import string
import numpy as np
try:
    from sklearn.metrics import classification_report
except Exception as e:
    print("Try ` pip install scikit-learn `")
    pass 
from tqdm import tqdm 

from .data import load_json

def classification_matrix(gold_ids, pred_ids, label_names, save_file):
    """
        gold_ids = [0,1,2]
        pred_ids = [1,1,2]
        label_names = ['a','b','c']
    """
    report = classification_report(gold_ids, pred_ids, labels=list(range(len(label_names))), target_names=label_names)
    with open(save_file, 'w', encoding='utf8') as f:
        f.write(report)


def normalize_answer(s):
    """Lower text and remove punctuation, articles and extra whitespace."""

    def remove_articles(text):
        regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
        return re.sub(regex, " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    if s == '':
        return s
    return white_space_fix(remove_articles(remove_punc(lower(s))))


def get_tokens(s):
    if not s:
        return []
    return normalize_answer(s).split()


def compute_f1(a_gold, a_pred):
    gold_toks = get_tokens(a_gold)
    pred_toks = get_tokens(a_pred)
    common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
    num_same = sum(common.values())
    if len(gold_toks) == 0 or len(pred_toks) == 0:
        # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
        return int(gold_toks == pred_toks)
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_toks)
    recall = 1.0 * num_same / len(gold_toks)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def prec_at_k(output, target, top_k=(1,)):
    """Computes the precision@k for the specified values of k
        Taken from  https://github.com/lyakaap/pytorch-template
    """
    max_k = max(top_k)
    batch_size = target.size(0)

    _, pred = output.topk(max_k, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in top_k:
        correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))

    if len(res) == 1:
        res = res[0]

    return res



if __name__ == '__main__':
    f1 = compute_f1("你好？hh", "NO go away 不好")
    f1 = compute_f1("你好？", "你好")
    print(f1)
    a = [1,1,1,0,0]
    b = [1,1,0,0,0]

    gold_ids = [0,1,2]
    pred_ids = [1,1,2]
    label_names = ['a','b','c']
    classification_matrix(gold_ids, pred_ids, label_names, 'tmp.txt')



from argparse import Namespace
from qdls.data import load_json, save_json
from qdls.utils import print_string
import pandas as pd
import datasets 
from tqdm import tqdm
from collections import defaultdict
from neo4j import GraphDatabase

from qdls.kgqa_eval.evaluator.base import BaseEvaluator
from qdls.kgqa_eval.evaluator.metric_fns import calc_metrics_per_sample, metric_fns


def get_answer(query, neo4j_config):
    driver = GraphDatabase.driver(uri=neo4j_config.neo4j_uri, auth=(neo4j_config.neo4j_user, neo4j_config.neo4j_passwd))
    with driver.session() as session:
        result = session.run(query)
        data = result.data()
    return sum([list(d.values()) for d in data], [])

def unwind_list(nested_list):
    result = []
    for item in nested_list:
        if isinstance(item, list):
            result.extend(unwind_list(item))
        else:
            result.append(str(item))
    return result

@metric_fns.register("metaqa_acc")
def metaqa_acc(sample, **kwargs):
    gold= sample['answers']
    try:
        answers = get_answer(sample['pred'], neo4j_config=kwargs['neo4j_config']) # 从neo4j中查询
    except: # 语法有误/ 执行错误直接视为[]
        answers = []
    pred = [_ for _ in answers if _ is not None]
    pred = unwind_list(pred)
    gold = unwind_list(gold)
    if set(gold).issubset(set(pred)):
        return True
    else:
        return False 
    

class MetaqaEval(BaseEvaluator):
    def __init__(self, file=None, lang='cypher', nproc=1, **kwargs) -> None:
        self.lang = lang
        self.nproc = nproc
        self.neo4j_config = kwargs.get('neo4j_config', "please set neo4j_config in kwargs")
        if kwargs.get("metrics_to_calc", None) is not None:
            self.metrics_to_calc = kwargs['metrics_to_calc']
        else:
            self.metrics_to_calc = ['bleu', 'executable', 'exact_match', "metaqa_acc"]
        # if lang == 'cypher':
        #     self.metrics_to_calc.extend(['is_correct'])

        print_string(f"metrics to calc: {self.metrics_to_calc}")
        print_string(f"neo4j config: {self.neo4j_config}")

        if file is not None:
            self.raw_data = load_json(file) if type(file) is str else file   
            for i, s in enumerate(self.raw_data):
                s['sample_id'] = i
            self.data_to_eval = [ self.normalize_sample(s, lang) for s in self.raw_data ]
            self.__pre_calc_metrics()
            self.id2sample = {sample['sample_id']:sample for sample in self.data_to_eval}
    
    def evaluate(self, verbose=True):
        scores = self.calc_marco_metrics()
        if verbose:
            for k,v in scores.items():
                print(f"{k:<20}: {v:.4f}")
        return scores
    
    def __pre_calc_metrics(self):
        """ 使用datasets库 并行计算 """
        ds = datasets.Dataset.from_list(self.data_to_eval)
        print(ds)
        
        if self.nproc > 1:
            ds = ds.map(
                calc_metrics_per_sample, num_proc=self.nproc, 
                fn_kwargs={
                    'ref_key':self.lang, 
                    'metrics':self.metrics_to_calc,
                    'neo4j_config': self.neo4j_config
                }
            )
            self.data_to_eval = ds.to_list()
        else:
            self.data_to_eval = [] 
            for s in tqdm(ds):
                self.data_to_eval.append(
                    calc_metrics_per_sample(s, self.lang, self.metrics_to_calc, self.neo4j_config)
                )

    def calc_marco_metrics(self, ids=None):
        """ 根据预先计算好的metrics, 为ids中的样本计算平均值 """
        if ids is None:
            ids = self.id2sample.keys()
        L = len(ids)
        scores = defaultdict(list)
        for i in ids:
            sample = self.id2sample[i]    
            for m in self.metrics_to_calc:
                scores[m].append(float(sample.get(m, 0)))
        return {k:sum(v)/L for k,v in scores.items()}
    
    def save_with_results(self, target_path):
        save_json(self.data_to_eval, target_path)  

    
class MetaqaCausalEval(MetaqaEval):
    @staticmethod
    def _process_raw(query):
        cql = query.split('[query]')[-1]
        return cql.strip()
    
    def normalize_sample(self, sample, lang=None):
        d = {
            'sample_id': sample['sample_id'],
            'question': sample['question'],
            'answers': sample['answers'],
            'cypher': sample['cypher'],
            'pred': MetaqaCausalEval._process_raw(sample['pred']),
        }
        return d

class MetaqaSeq2seqEval(MetaqaEval):
    def normalize_sample(self, sample, lang=None):
        pred = sample['pred'].strip()
        d = {
            'sample_id': sample['sample_id'],
            'question': sample['question'],
            'answers': sample['answers'],
            'cypher': sample['cypher'],
            'pred': pred,
        }
        return d




import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import confusion_matrix
from matplotlib.font_manager import FontProperties  # 字体管理器
import random


def plot_confusion_matrix(true_label, pred_label, save_path=None):
    # 设置汉字格式
    font = FontProperties(size=12)
    
    sns.set()
    f, ax = plt.subplots()
    C2 = confusion_matrix(true_label, pred_label, labels=[0, 1])
    sns.heatmap(C2, annot=True, ax=ax)  # 画热力图
    
    plt.title("混淆矩阵", fontproperties=font)  # 设置标题
    plt.xlabel('预测值', fontproperties=font)  # 设置x轴标题
    plt.ylabel('真实值', fontproperties=font)  # 设置y轴标签
    
    # 保存绘图
    if save_path is not None:
        plt.savefig(save_path)
    
    # 绘图
    plt.show()
    

def multiple_line_plot( labels, value_lists, save_path=None):
    """绘制多条折线图

    Args:
        labels: str list. 每条折线的标签
        value_lists: list of list. 每条折线的值
        save_path: 图片保存路径. Defaults to None.
    """
    ls_options = ['-', '--', '-.', ':', 'None', ' ', '', 'solid', 'dashed', 'dashdot', 'dotted']
    for label, value_list in zip(labels, value_lists):
        plt.plot(value_list, marker="o", label=label, ls="--")
        # plt.plot(value_list, marker="o", label=label)
  
    
    plt.xlabel("month")
    plt.ylabel("income")
    plt.title("income of each Company")
    
    plt.grid(color='lightpink', linestyle='--', linewidth=0.5)
    plt.legend()
    plt.show()
    
    if save_path is not None:
        plt.savefig(save_path)


if __name__ == "__main__":
    def test_confusion_matrix():
        true_label = [random.randint(0, 1) for i in range(0, 100)]
        pred_label = [random.randint(0, 1) for i in range(0, 100)]
        plot_confusion_matrix(true_label, pred_label)
                
    def test_multiple_line_plot():
        label_name = ["A", "B", "C", "D"]
        metric_name = "income"
        lists = [[random.randint(0, 100) for i in range(0, 12)] for i in range(0, 4)]
        multiple_line_plot(label_name, lists)
        
    
    test_multiple_line_plot()


# -*- coding: utf-8 -*-
# @File    :   loggers.py
# @Time    :   2023/03/26 10:43:18
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    封装一下常用的logger
'''
from pytorch_lightning.loggers import TensorBoardLogger, WandbLogger


def get_wandb_logger(config):
    
    logger = WandbLogger(project=config.project, name=config.version, job_type=config.job_type)

    return logger 


def get_tfb_logger(config):
    logger = TensorBoardLogger(
        save_dir="./lightning_logs/",
        name=None,                # 指定experiment, ./lightning_logs/exp_name/version_name
        version=config.version,   # 指定version, ./lightning_logs/version_name
    )
    return logger


# -*- coding: utf-8 -*-
# @File    :   examples.py
# @Time    :   2024/06/26 15:14:02
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
    调用 API 示例
'''
from loguru import logger
import os 
try:
    import chattool
except Exception as e:
    logger.warning(f"chattool not found: {e}, run `pip install -U chattool`")
    

TONGYI_API_KEY = os.environ.get('TONGYI_API_KEY', None) 
TONGYI_API_URL = "https://dashscope.aliyuncs.com/compatible-mode/v1"

CHATANY_API_KEY = os.environ.get('CHATANY_API_KEY', None)
CHATANY_API_URL = os.environ.get('CHATANY_API_URL' ,"https://api.chatanywhere.tech/v1")

KIMI_API_KEY = os.environ.get('KIMI_API_KEY', None)
KIMI_API_URL = os.environ.get('KIMI_API_URL', "https://api.moonshot.cn/v1")


ZHIPU_API_KEY = os.getenv("ZHIPU_API_KEY", None)

def openai_demo(prompt):
    """Openai 官方库调用通义API示例
        json format 的使用示例
    Args:
        prompt: str
    """
    from openai import OpenAI

    client = OpenAI(
        api_key=TONGYI_API_KEY,  # 替换成真实DashScope的API_KEY
        base_url="https://dashscope.aliyuncs.com/compatible-mode/v1",  # 填写DashScope服务endpoint
    )

    completion = client.chat.completions.create(
        model="qwen-long",
        messages=[
            {
                'role': 'system',
                'content': 'You are a helpful assistant.'
            },
            {
                'role': 'user',
                'content': prompt
            },

        ],
        stream=False,
        temperature=0.0,
        # response_format={ "type": "json_object" } # 返回json格式式 prompt 中要有 json 这个关键字
    )
    print(completion)
    # print(completion.choices[0].message.content)
    return completion



def zhipu_demo():
    """ 
    https://open.bigmodel.cn/dev/api#language
    """
    from zhipuai import ZhipuAI
    client = ZhipuAI(api_key=ZHIPU_API_KEY)
    response = client.chat.completions.create(
    model="glm-4",  # 填写需要调用的模型名称
    messages=[
            {"role": "user", "content": "作为一名营销专家，请为智谱开放平台创作一个吸引人的slogan"},
            {"role": "assistant", "content": "当然，为了创作一个吸引人的slogan，请告诉我一些关于您产品的信息"},
            {"role": "user", "content": "智谱AI开放平台"},
            {"role": "assistant", "content": "智启未来，谱绘无限一智谱AI，让创新触手可及!"},
            {"role": "user", "content": "创造一个更精准、吸引人的slogan"}
        ],
    )
    print(response.choices[0].message)
    return  

def kimi_demo():
    """
    https://platform.moonshot.cn/docs/intro#%E4%B8%BB%E8%A6%81%E6%A6%82%E5%BF%B5 
    兼容 Openai API
    "moonshot-v1-8k"
    """
    pass


def chattool_demo(prompt="hello"):
    import chattool
    API_URL = "https://api.chatanywhere.tech/v1"
    API_KEY = os.environ.get('CHATANY_API_KEY', None) # local vllm set to "" 
    chattool.api_base = API_URL
    chattool.api_key = API_KEY
    c = chattool.Chat(model="gpt-4-turbo")
    # c.system()
    c.user(prompt)
    c.getresponse(
        temperature=0.0, 
        response_format={ "type": "json_object" }
        )
    print(c.chat_log)
    return c 

def chattool_parallel_demo():
    import chattool

    chattool.base_url=CHATANY_API_URL; chattool.api_key = CHATANY_API_KEY
    prompts = [] 
    model = 'gpt-4o-mini'; version= 'test_demo'
    ckpt_path = f"./cache/{model}_{version}.jsonl"
    def data2chat(prompt):
        c = chattool.Chat()
        c.user(prompt)
        return c

    chattool.async_chat_completion(
        prompts,
        model=model,
        chkpoint=ckpt_path,
        nproc=8,
        data2chat=data2chat,
        timeinterval=1, max_tries=3, timeout=3,
        stop="</s>"
    )
    chats = chattool.load_chats(ckpt_path)

    print(len(chats), "loaded! ")

    for chat in chats:
        output = chat.chat_log[-1]['content']


def langchain_demo(prompt="hello"):
    """ 
     langchain OpenAI 调用本地模型
    """
    from langchain_openai import ChatOpenAI

    llm = ChatOpenAI(
        model="/sshfs/pretrains/THUDM/glm-4-9b-chat",
        temperature=0,
        openai_api_key='EMPTY',
        openai_api_base=f'http://127.0.0.1:8000/v1', # sparql only proto
        max_tokens=1024,
        verbose=True
    )
    # to pass stop_token_ids to vllm 
    extra_body={
        "stop_token_ids": [151329, 151336, 151338]
        }
    print(llm.invoke(prompt, extra_body=extra_body))

if __name__ == '__main__':
    from langchain_core.messages import HumanMessage, AIMessage, SystemMessage
    prompt = [SystemMessage(), HumanMessage("你好")]
    langchain_demo(prompt)
    pass 






import os
import sys
import json
import yaml
import argparse
from  argparse import Namespace
import omegaconf
from omegaconf import OmegaConf as oc

import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
import multiprocessing
from colorama import Fore, Back, Style
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

# from tensor_ops import safesoftmax
# from q_snippets.tensor_ops import safesoftmax

from abc import ABC, abstractmethod
from pytorch_lightning.loggers import TensorBoardLogger


class RoPE(pl.LightningModule):
    def __init__(self, dim, max_seq_len=513, device=None):
        """
            ref: https://kexue.fm/archives/8265
            创建最大长度，维度固定的pos_emebddings
        """
        super().__init__()
        self.dim = dim
        self.seq_len = max_seq_len
        # self.device = device
        self.embeddings = self._create_pos_embedding()

    def _create_pos_embedding(self):
        position_ids = torch.arange(0, self.seq_len, dtype=torch.float, device=self.device)[None]  # (1, seqlen)
        indices = torch.arange(0, self.dim//2, dtype=torch.float, device=self.device)  # odd not work?
        indices = torch.pow(10000.0, -2*indices/self.dim)
        
        embeddings = torch.einsum("bn,d->bnd", position_ids, indices)
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = torch.reshape(embeddings, (-1, self.seq_len, self.dim))
        return embeddings
    
    def add_pos_embedding(self,qw):
        """
            输入向量序列(bsz,seq_len,dim)，返回乘上了RoPE的结果
        """
        bsz, seq_len, dim = qw.size()
        pos = self.embeddings[:, seq_len, :]
        cos_pos = torch.repeat_interleave(pos[..., 1::2], 2, dim=-1).to(self.device)
        sin_pos = torch.repeat_interleave(pos[..., ::2], 2, dim=-1).to(self.device)
        qw2 = torch.stack([-qw[...,1::2], qw[...,::2]], dim=-1)
        qw2 = torch.reshape(qw2, qw.shape)
        # print(self.device, qw.device, qw2.device, cos_pos.device)
        qw = qw*cos_pos + qw2*sin_pos
        return qw


class TrainerProcess(multiprocessing.Process, ABC):
    """ Only for Cross validation """
    def __init__(self, config, module, dm, k=None):
        """

        Args:
            config (Config):  配置信息
            module (LightningModule):  要运行的 LightningModule
            dm (LightningDataModule):     DataModule
            k (int, optional): 第几个fold. Defaults to None.
        """
        super().__init__()
        self.k = k
        config.sub_dir = os.path.join(config.version, f"fold_{k}" if k is not None else k) 
        self.config = self._prepare_config(config)
        self.model = module(config)
        self.dm = dm
        
        self.logger = self._set_logger()
        self.callbacks = self._set_callbacks()
        self.trainer = self._set_trainer()
    
    @property
    def dirname(self):
        return os.path.join(os.getcwd(), "lightning_logs", self.config.sub_dir)

    @abstractmethod
    def _prepare_config(self, config):
        # logger and model_ckpt require same directory
        
        return config

    @abstractmethod
    def _set_callbacks(self):
        """设置保存模型的路径及参数, 第一个最好是ModelCheckPointCallback
        set self.callbacks = 
            a list of callbacks
        """
        pass
        
    @abstractmethod
    def _set_trainer(self):
        """set self.trainer = 
        """
        pass


    def _set_logger(self):
        """set self.logger here
        """
        logger = TensorBoardLogger(
            save_dir="./lightning_logs/",
            name=None,    # 指定experiment, ./lightning_logs/exp_name/version_name
            version=self.dirname,    # 指定version, ./lightning_logs/version_name
        )
        return logger
        
        
    def run_train(self):
        """进程执行的主函数

        """
        self.trainer.fit(self.model, self.dm)
        # return self.callbacks[0].best_model_path
        self.callbacks[0].to_yaml(os.path.join(self.dirname, "best_k_path.yaml"))

    def run(self):
        self.run_train()


class RNNWrapper(nn.Module):
    """
    Args:
        rnn (nn.Module):  RNN to run 
    """
    def __init__(self, rnn):
        super().__init__()
        self.rnn = rnn
        
    def forward(self, input_sequence, attention_mask):
        """
            输出dim与RNN期待的输出维度相同
        Args:
            input_sequence (Tensor): (bsz,seq_len,dim)
            attention_mask (Tensor): (bsz,seq_len),  0 for masking

        Returns:
            tuple : output => (bsz,seq_len,dim),  padding的位置是0向量
                    hidden => (num_layers*bidirectional, bsz, dim) 如果是LSTM有两个:(last_hidden,cell_state) 不是batch_first
        """
        pack = pack_padded_sequence(input_sequence, attention_mask.sum(1), batch_first=True, enforce_sorted=False)
        o, h = self.rnn(pack)
        output, seq_lens = pad_packed_sequence(o, batch_first=True)
        return output, h


class MultiHeadAttnWrapper(nn.Module):
    def __init__(self, dim, num_heads):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.atten = nn.MultiheadAttention(dim, num_heads)
        
    def forward(self, q, k, v, q_mask=None, k_mask=None):
        """  1 (bool()转成True)的位置被mask掉

        Args:
            q (Tensor): (bsz,seqlen_target,dim)
            k (Tensor): (bsz,seqlen_source,dim)
            v (Tensor): (bsz,seqlen_source,dim)
            q_mask (Tensor, optional):  (bsz, seqlen_target, optional dim), 1 for masking. Defaults to None.
            k_mask (Tensor, optional): (bsz, seqlen_source, optional dim), 1 for masking. Defaults to None.

        Returns:
            tuple:
                attended_q         (bsz, seqlen_target, dim)    q中的每个向量都是v的加权求和
                attention_weight   (bsz, seqlen_target, seqlen_source)
        """
        if q_mask is not None and k_mask is not None:
            q_mask = q_mask[...,None] if len(q_mask.size()) == 2 else q_mask
            k_mask = k_mask[...,None] if len(k_mask.size()) == 2 else k_mask
            try:
                mask = q_mask.long() & k_mask.permute((0,2,1)).long()              # (bsz, seqlen_target, seqlen_source)
            except:
                mask = q_mask @ k_mask.permute((0,2,1))
            mask = mask.repeat_interleave(self.num_heads, dim=0).bool()
        else:
            mask = None
        atten, attention_weight = self.atten(
            q.permute((1,0,2)),                       # (seqlen_target, bsz, dim)
            k.permute((1,0,2)),v.permute((1,0,2)),    # (seqlen_source, bsz, dim)
            attn_mask=mask                            # (bsz*nheads, seqlen_target, seqlen_source)
        )
        return atten.permute((1,0,2)), attention_weight

# class Config(argparse.Namespace):
class Config(oc):
    def __init__(self):
        pass

    def __repr__(self):
        return json.dumps(self.__dict__, indent=2, default=lambda x: x.__dict__, ensure_ascii=False) 

    def as_dict(self):
        return eval(json.dumps(self.__dict__, indent=2, default=lambda x: x.__dict__, ensure_ascii=False))

    def save_to_json_file(self, to_file):
        with open(to_file, "w") as fout:
            json.dump(self.__dict__, fout, indent=2, default=lambda x: x.__dict__, ensure_ascii=False)
            print(f"----- config saved to {to_file} -------")

    @classmethod
    def from_json_file(cls, in_file):
        with open(in_file) as fin:
            config_dict = json.load(fin)
            return cls.from_dict(config_dict)

    @classmethod
    def from_yaml_file(cls, in_file):
        """ 可以直接使用Config.load(yaml_path), 调用的是OmegaConf的load """
        try:
            config_dict = yaml.safe_load(open(in_file))
        except:
            config_dict = yaml.load(open(in_file), Loader=yaml.Loader)
        return cls.from_dict(config_dict)

    @classmethod
    def from_dict(cls, config_dict):
        config = cls()
        to_remove = []
        for key, value in config_dict.items():
            if isinstance(value, dict):
                value = cls.from_dict(value)
            if hasattr(config, key):
                setattr(config, key, value)
                to_remove.append(key)
            else:
                setattr(config, key, value)

        if to_remove:
            print("Config override keys: %s" % ",".join(to_remove))

        return config       
    
    def save_to_yaml_file(self, to_file):
        """保存OmegaConf到yaml文件

        Args:
            config (OmegaConf): config 对象
            path (str): 目标yaml
        """
        with open(to_file, 'w') as f:
            config = oc.create(self.as_dict())            
            oc.save(config, f)
            print(f"----- config saved to {to_file} -------")

    @staticmethod
    def from_argparse(parser: argparse.ArgumentParser):
        """返回从
            一个ArgumentParser对象获取的命令行参数，
            没出现在命令行中设定好默认值的参数
           两个OmegaConf对象。
        ！命令行设置时，不要加=， 而是直接空格隔开 !! 可以用=了

        Args:
            parser (argparse.ArgumentParser): 入口文件中定义的parser

        Returns:
            tuple: (OmegaConf, OmegaConf)
        """
        #  {'data.train_bsz': '--train_bsz'}
        dest_to_arg = {v.dest: k for k, v in parser._option_string_actions.items()}

        all_args = vars(parser.parse_args())
        provided_args = {}
        default_args = {}
        sys_argv = [ s.split("=")[0] for s in sys.argv[1:]]  # 除去 =
        for k, v in all_args.items():
            if dest_to_arg[k] in sys_argv:
                provided_args[k] = v
            else:
                default_args[k] = v
        # print(provided_args,"\n---------------\n", default_args)
        provided_dotlist = [ f"{k}={v}" for k,v in provided_args.items()] if provided_args != {} else []
        provided = oc.from_dotlist(provided_dotlist)
        default_dotlist = [ f"{k}={v}" for k,v in default_args.items()]
        defaults = oc.from_dotlist(default_dotlist)

        return provided, defaults  


# def print_string(string):
#     """ 两边是==，输入字符串居中，两边各有5个空格，输出字体颜色为黄色 """
#     print(Fore.YELLOW,"="*25 + string.center(len(string)+10,) + "="*25)
#     print(Style.RESET_ALL)

# def print_config(config):
#     """ 打印字典，config等， 字体颜色为绿色 """
#     print("="*80)
#     if type(config) is omegaconf.dictconfig.DictConfig:
#         config_dict = oc.to_object(config)
#     else:
#         config_dict = namespace2dict(config)
#     s = json.dumps(config_dict, ensure_ascii=False, indent=2)
#     print(Fore.GREEN, s)
#     print(Style.RESET_ALL)
#     print("="*80)

def resume_config(config, path):
    """ 
    pytorch lightning save_hyperparameters() 保存的yaml文件 在预测的时候需要更新
    从原先保存的yaml恢复，重写新的 mode max_epochs, data and preds 
        config: 当前的config
        path: yaml 路径

    """

    new_config = Config.load(path)
    # new_config = oc.merge(old_config, config)
    new_config.mode = config.mode 
    new_config.max_epochs = config.max_epochs

    new_config.data.test_path = config.data.test_path
    new_config.data.test_bsz = config.data.test_bsz
    new_config.data.force_reload = config.data.force_reload

    new_config.preds.ckpt_path = config.preds.ckpt_path
    new_config.preds.result_path = config.preds.result_path
    print_string(f"config file resumed from {path}")
    return new_config



def build_config(d):
    """ 嵌套的 argparse.Namespace 
        TODO: move to utils
    """
    config = Namespace()
    for k,v in d.items():
        if type(v) == dict:
            setattr(config,k, build_config(v))
        else:
            setattr(config, k, v)
    return config 
    
def namespace2dict(config):
    """
     TODO: move to utils, rename to config2dict
    """
    d = {}
    for k,v in config.__dict__.items():
        if type(v) is Namespace:
            d[k] = namespace2dict(v)
        else:
            d[k] = v 
    return d 


if __name__ == '__main__':
    # print( os.getcwd())
    # config = Config.from_dict({'a':23, 'b':{'c':'we'}})
    # print(config)
    # config.save_to_json_file("tmp.json")
    # c1 = Config.from_json_file("tmp.json")
    # print("c1", c1)
    # c1.save_to_yaml_file("tmp.yaml")
    # c2 = Config.from_yaml_file("tmp.yaml")
    # print("c2", c2)
    # print(hasattr(c2, 'model_specific'))
    # c2.model_specific = {'k':10, 'n':50}
    # print(hasattr(c2, 'model_specific'), c2)
    # print_info("This config:")


    # c2 = Config.load("./tmp.yaml")
    # print_config(c2)
    # print_string("THis is good"*5)
    
    def test_namespace():
        d = {
            'name': {
                '1': 1,
                '2':2
            },
            'list': [1,2,3]
        }
        c = build_config(d)
        print(c)
        print(namespace2dict(c))

    test_namespace()



import os 
import json 
from tqdm import tqdm 
from abc import ABC, abstractmethod

from qdls.data import load_json, save_json, load_jsonl
from vllm import LLM, SamplingParams

class VLLMBaseRunner:
    """ 
    config: OmegaConf
        version: str
        model_path: str
        temperature: float
        max_tokens: int
        stop: str
        tensor_parallel_size: int

        实现build_prompt_fn
        cache 文件夹是生成缓存
        prompts 文件夹是 prompt 缓存, 有些 prompt 生成速度较慢
    """
    def __init__(self, config) -> None:
        self.config = config 

        self.version = config.version 
        self.cache_file = self._handle_cache()
        tensor_parallel_size = getattr(config, 'tensor_parallel_size', 1)
        stop_token = getattr(config, 'stop', '</s>')

        self.test_data = load_json(config.test_path)
        self.prompts = self._handle_prompt_cache()

        self.llm = LLM(
            model=config.model_path,
            tensor_parallel_size=tensor_parallel_size,
            trust_remote_code=True
        )
        

        self.sampling_params = SamplingParams(
                temperature=config.temperature, 
                max_tokens=config.max_tokens, 
                stop=stop_token
            )


    def _handle_cache(self):
        """ define cache file name """
        if not os.path.exists("./cache/results"):
            os.makedirs("./cache/results")
        self.cache_file = f"./cache/results/{self.version}.jsonl"


        # 从缓存中加载 TODO: 目前没有用，不是调 API 计费的情况下，不太需要缓存
        if os.path.exists(self.cache_file):
            print(f"cache file {self.cache_file} exists, loading...")
            self.processed = load_jsonl(self.cache_file)
            print(f"loaded {len(self.processed)} chats")

        return open(self.cache_file, 'a')


    def _handle_prompt_cache(self):
        if not os.path.exists("./cache/prompts"):
            os.makedirs("./cache/prompts", exist_ok=True)

        self.prompt_cache_file = f"./cache/prompts/{self.version}.json"

        if os.path.exists(self.prompt_cache_file):
            print(f"cache file {self.prompt_cache_file} exists, loading...")
            prompts = load_json(self.prompt_cache_file)
            print(f"loaded {len(prompts)} chats")
        else:
            prompts = self.build_prompts(self.test_data)
            save_json(prompts, self.prompt_cache_file)

        return prompts

    @abstractmethod
    def build_prompt_fn(sample):
        """ 为一条数据构建prompt 
            应该是一个 staticmethod
        """
        pass 


    def build_prompts(self, samples, nproc=32):
        """ if use_tqdm 是数据量非常多的情况下，可以使用tqdm显示进度条; 否则是为一个batch构建prompt """

        if nproc == 1:
            R = [] 
            for sample in tqdm(samples, desc="Building prompts"):
                prompt = self.build_prompt_fn(sample)
                R.append(prompt)
            return R 
        else:
            from multiprocessing import Pool 
            with Pool(nproc) as p:
                R = list(tqdm(p.imap(self.build_prompt_fn, samples), total=len(samples), desc=f"Building prompts with {nproc} processes"))
            return R
        


    def generate_from_samples(self, data, bsz=8):
        """ 
        从原始的数据自行分batch进行生成，生成一个 batch 存储一下，如此 cache file 才有意义
        不足：构建 prompt 比较耗费时间的话，这样会导致生成速度变慢
        Args:
            data: list of samples 
            bsz: batch size
        
        """

        for i in tqdm(range(len(data)//bsz+1)):
            batch = data[i*bsz:(i+1)*bsz]
            if batch == []:
                break

            prompts = self.build_prompts(batch)

            outputs = self.llm.generate(prompts, self.sampling_params, use_tqdm=False)

            # Print the outputs.
            for sample, output in zip(batch, outputs):
                prompt = output.prompt
                generated_text = output.outputs[0].text
                sample['vllm'] = generated_text 

                self.cache_file.write(json.dumps(sample, ensure_ascii=False) + '\n')

        self.cache_file.close()


    def generate_from_prompts(self):
        """ 先构建好所有的prompts 交给vllm生成 
            生成完之后，再存储
        """
 
        outputs = self.llm.generate(self.prompts, self.sampling_params, use_tqdm=True)
        for sample, output in zip(self.test_data, outputs):
            generated_text = output.outputs[0].text
            sample['vllm'] = generated_text
            sample['prompt'] = output.prompt
            self.cache_file.write(json.dumps(sample, ensure_ascii=False) + '\n')
        
        self.cache_file.close()

        




from abc import abstractmethod, ABC

class Retriever(ABC):
    
    @abstractmethod
    def get_topk_samples(self, query, topk=5):
        """ 
            return topk samples according to query
        """
        raise NotImplementedError
    





# -*- coding: utf-8 -*-
# @File    :   num_params.py
# @Time    :   2024/04/06 15:18:43
# @Author  :   Qing 
# @Email   :   aqsz2526@outlook.com
######################### docstring ########################
'''
'''
import fire 

from transformers import AutoModel

def num_params(model_path):
    """ 
    model_path: str, huggingface 模型的路径
    Return: int, 模型的参数量(浮点数的个数)
    """
    model = AutoModel.from_pretrained(model_path, trust_remote_code=True)
    total_params = sum(p.numel() for p in model.parameters())
    print(f"模型 {model_path} 的参数量为：{total_params}, {total_params/1000000000:.2f} B, {total_params/1000000:.2f} M")
    return total_params


if __name__ == '__main__':
    fire.Fire(num_params)




from abc import ABC, abstractmethod
from typing import List, Any,  Union, Callable
from jieba import cut
import os 
import pickle
from loguru import logger
from rank_bm25 import BM25Okapi
from qdls.gql.cypher.utils.parse import split_cypher # TODO: unified interface for cypher sparql, etc 

from qdls.llm_infer.retrievers.base_retriever import Retriever

class LanguageProcessor(ABC):
    @abstractmethod
    def tokenize(self, document: str) -> List[str]:
        pass

class EnglishProcessor(LanguageProcessor):
    def tokenize(self, document: str) -> List[str]:
        return document.split(" ")

class ChineseProcessor(LanguageProcessor):
    def tokenize(self, document: str) -> List[str]:
        return list(cut(document))

class CypherProcessor(LanguageProcessor):
    def tokenize(self, document: str) -> List[str]:
        return split_cypher(document)

class BM25Retriever(Retriever):
    def __init__(self, data_source: List[Any], key: Union[str, Callable] = None, lang: str = 'en', **kwargs) -> None:
        """
        初始化一个处理文本数据的实例。

        参数:
        - data_source: List[Any], 数据源列表，每个元素可以是任何类型，但必须包含可处理的文本信息。
        - key: str, 用于从数据源中提取文本信息的键。此键应指向包含文本数据的字段。
        - lang: str, 文本的语言，默认为'en'表示英语。
        - kwargs: dict, 其他参数。 一般不需要使用，可以传入cache_file参数指定缓存文件名(./cached_index/{cache_file})。

        返回值:
        - 无
        """
        # 确保key已经被指定
        assert key is not None, "key is not specified"
        self.key = key 
        # 确保data_source是一个列表
        assert isinstance(data_source, list), "data_source must be a list"
        self.data_source = data_source
        # 使用指定的key处理数据源，提取需要的文本项
        self.items4retrieval = self._process_key(key)
        self.lang = lang
        # 初始化语言处理器，用于处理特定语言的文本
        self.language_processor = self._get_language_processor()

        self.cache_file = kwargs.get("cache_file", None)
        self._handle_cache()

    def _process_key(self, key: Union[str, Callable]) -> List[Any]:
        """
        根据提供的键对数据源中的每个样本进行处理。
        
        如果键是一个可调用对象，则将对每个样本应用该函数，并返回结果列表。
        如果键是一个字符串，则将它作为样本字典的键来访问，并返回每个键对应的值的列表。
        
        参数:
        key: str - 用于处理数据源样本的键。可以是函数，也可以是字符串。
        
        返回值:
        List[Any] - 根据提供的键处理数据源样本后的结果列表。
        """
        if callable(key):
            # 如果key是可调用的，则对每个样本应用该函数
            return [key(sample) for sample in self.data_source]
        elif isinstance(key, str):
            # 如果key是字符串，则用它来访问每个样本的值
            return [sample[key] for sample in self.data_source]
        else:
            # 如果key既不是可调用的也不是字符串，则抛出异常
            raise Exception(f"key should be dict key or callable function")

    def _get_language_processor(self) -> LanguageProcessor:
        if self.lang == 'en':
            return EnglishProcessor()
        elif self.lang == 'zh':
            return ChineseProcessor()
        elif self.lang == 'cypher':
            return CypherProcessor()
        else:
            raise ValueError(f"lang must be 'en', 'zh', or 'cypher', but got {self.lang}")

    def _tokenize_data(self) -> List[List[str]]:
        return [self.language_processor.tokenize(doc) for doc in self.items4retrieval]

    def get_topk_samples(self, query: str, topk: int = 5) -> List[Any]:
        """
        获取与查询字符串最相关的前k个样本。
        
        参数:
        - query: str，用户输入的查询字符串。
        - topk: int，返回的最相关样本数量，默认为5。
        
        返回值:
        - List[Any]，包含最相关样本的列表。
        """
        # 对查询字符串进行分词处理
        tokenized_query = self.language_processor.tokenize(query)
        # 使用BM25算法从数据源中获取与分词后的查询最相关的前topk个样本
        topk_samples = self.bm25.get_top_n(tokenized_query, self.data_source, n=topk)
        return topk_samples
    
    def _handle_cache(self):
        """ 
            将 self.bm25 缓存到文件中
        """
        os.makedirs("cached_index", exist_ok=True)
        if self.cache_file is  None:
            self.cache_file = f"{ self.key if type(self.key) is str else self.key.__name__ }"
        cache_file = os.path.join(".", "cached_index", f"{self.cache_file}_bm25_retriever.pkl")
        self.cache_file = cache_file

        if os.path.exists(cache_file):
            with open(cache_file, "rb") as f:
                self.bm25 = pickle.load(f)
                logger.info(f"BM25 index loaded from {cache_file}")
        else:
            # 对数据源进行分词处理
            self.tokenized_corpus = self._tokenize_data()
            # 初始化BM25索引，用于文本的检索
            self.bm25 = BM25Okapi(self.tokenized_corpus)
            with open(cache_file, "wb") as f:
                pickle.dump(self.bm25, f)
                logger.info(f"BM25 index saved to {cache_file}")


if __name__ == '__main__':
    data = [
        {
            'question': "asdfa",
            "answer": "sdfa"
        },

    ]

    bm25 = BM25Retriever(data, key='question', lang='zh')
    datalist = bm25.get_topk_samples("asdfa", topk=1)